# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_discovery_v1.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v5.protos import ydb_discovery_pb2 as protos_dot_ydb__discovery__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16ydb_discovery_v1.proto\x12\x10Ydb.Discovery.V1\x1a\x1aprotos/ydb_discovery.proto2\xb5\x01\n\x10\x44iscoveryService\x12Z\n\rListEndpoints\x12#.Ydb.Discovery.ListEndpointsRequest\x1a$.Ydb.Discovery.ListEndpointsResponse\x12\x45\n\x06WhoAmI\x12\x1c.Ydb.Discovery.WhoAmIRequest\x1a\x1d.Ydb.Discovery.WhoAmIResponseBW\n\x1btech.ydb.proto.discovery.v1Z8github.com/ydb-platform/ydb-go-genproto/Ydb_Discovery_V1b\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_discovery_v1_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\033tech.ydb.proto.discovery.v1Z8github.com/ydb-platform/ydb-go-genproto/Ydb_Discovery_V1'
  _DISCOVERYSERVICE._serialized_start=73
  _DISCOVERYSERVICE._serialized_end=254
# @@protoc_insertion_point(module_scope)



from protos import ydb_operation_pb2 as _ydb_operation_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class StatusFlag(_message.Message):
    __slots__ = ()
    class Status(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        UNSPECIFIED: _ClassVar[StatusFlag.Status]
        GREY: _ClassVar[StatusFlag.Status]
        GREEN: _ClassVar[StatusFlag.Status]
        BLUE: _ClassVar[StatusFlag.Status]
        YELLOW: _ClassVar[StatusFlag.Status]
        ORANGE: _ClassVar[StatusFlag.Status]
        RED: _ClassVar[StatusFlag.Status]
    UNSPECIFIED: StatusFlag.Status
    GREY: StatusFlag.Status
    GREEN: StatusFlag.Status
    BLUE: StatusFlag.Status
    YELLOW: StatusFlag.Status
    ORANGE: StatusFlag.Status
    RED: StatusFlag.Status
    def __init__(self) -> None: ...

class SelfCheckRequest(_message.Message):
    __slots__ = ("operation_params", "return_verbose_status", "minimum_status", "maximum_level")
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    RETURN_VERBOSE_STATUS_FIELD_NUMBER: _ClassVar[int]
    MINIMUM_STATUS_FIELD_NUMBER: _ClassVar[int]
    MAXIMUM_LEVEL_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    return_verbose_status: bool
    minimum_status: StatusFlag.Status
    maximum_level: int
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., return_verbose_status: bool = ..., minimum_status: _Optional[_Union[StatusFlag.Status, str]] = ..., maximum_level: _Optional[int] = ...) -> None: ...

class SelfCheckResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class NodeCheckRequest(_message.Message):
    __slots__ = ("operation_params",)
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ...) -> None: ...

class NodeCheckResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class SelfCheck(_message.Message):
    __slots__ = ()
    class Result(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        UNSPECIFIED: _ClassVar[SelfCheck.Result]
        GOOD: _ClassVar[SelfCheck.Result]
        DEGRADED: _ClassVar[SelfCheck.Result]
        MAINTENANCE_REQUIRED: _ClassVar[SelfCheck.Result]
        EMERGENCY: _ClassVar[SelfCheck.Result]
    UNSPECIFIED: SelfCheck.Result
    GOOD: SelfCheck.Result
    DEGRADED: SelfCheck.Result
    MAINTENANCE_REQUIRED: SelfCheck.Result
    EMERGENCY: SelfCheck.Result
    def __init__(self) -> None: ...

class StoragePDiskStatus(_message.Message):
    __slots__ = ("id", "overall")
    ID_FIELD_NUMBER: _ClassVar[int]
    OVERALL_FIELD_NUMBER: _ClassVar[int]
    id: str
    overall: StatusFlag.Status
    def __init__(self, id: _Optional[str] = ..., overall: _Optional[_Union[StatusFlag.Status, str]] = ...) -> None: ...

class StorageVDiskStatus(_message.Message):
    __slots__ = ("id", "overall", "vdisk_status", "pdisk")
    ID_FIELD_NUMBER: _ClassVar[int]
    OVERALL_FIELD_NUMBER: _ClassVar[int]
    VDISK_STATUS_FIELD_NUMBER: _ClassVar[int]
    PDISK_FIELD_NUMBER: _ClassVar[int]
    id: str
    overall: StatusFlag.Status
    vdisk_status: StatusFlag.Status
    pdisk: StoragePDiskStatus
    def __init__(self, id: _Optional[str] = ..., overall: _Optional[_Union[StatusFlag.Status, str]] = ..., vdisk_status: _Optional[_Union[StatusFlag.Status, str]] = ..., pdisk: _Optional[_Union[StoragePDiskStatus, _Mapping]] = ...) -> None: ...

class StorageGroupStatus(_message.Message):
    __slots__ = ("id", "overall", "vdisks")
    ID_FIELD_NUMBER: _ClassVar[int]
    OVERALL_FIELD_NUMBER: _ClassVar[int]
    VDISKS_FIELD_NUMBER: _ClassVar[int]
    id: str
    overall: StatusFlag.Status
    vdisks: _containers.RepeatedCompositeFieldContainer[StorageVDiskStatus]
    def __init__(self, id: _Optional[str] = ..., overall: _Optional[_Union[StatusFlag.Status, str]] = ..., vdisks: _Optional[_Iterable[_Union[StorageVDiskStatus, _Mapping]]] = ...) -> None: ...

class StoragePoolStatus(_message.Message):
    __slots__ = ("id", "overall", "groups")
    ID_FIELD_NUMBER: _ClassVar[int]
    OVERALL_FIELD_NUMBER: _ClassVar[int]
    GROUPS_FIELD_NUMBER: _ClassVar[int]
    id: str
    overall: StatusFlag.Status
    groups: _containers.RepeatedCompositeFieldContainer[StorageGroupStatus]
    def __init__(self, id: _Optional[str] = ..., overall: _Optional[_Union[StatusFlag.Status, str]] = ..., groups: _Optional[_Iterable[_Union[StorageGroupStatus, _Mapping]]] = ...) -> None: ...

class StorageStatus(_message.Message):
    __slots__ = ("overall", "pools")
    OVERALL_FIELD_NUMBER: _ClassVar[int]
    POOLS_FIELD_NUMBER: _ClassVar[int]
    overall: StatusFlag.Status
    pools: _containers.RepeatedCompositeFieldContainer[StoragePoolStatus]
    def __init__(self, overall: _Optional[_Union[StatusFlag.Status, str]] = ..., pools: _Optional[_Iterable[_Union[StoragePoolStatus, _Mapping]]] = ...) -> None: ...

class ComputeTabletStatus(_message.Message):
    __slots__ = ("overall", "type", "state", "count", "id")
    OVERALL_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    overall: StatusFlag.Status
    type: str
    state: str
    count: int
    id: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, overall: _Optional[_Union[StatusFlag.Status, str]] = ..., type: _Optional[str] = ..., state: _Optional[str] = ..., count: _Optional[int] = ..., id: _Optional[_Iterable[str]] = ...) -> None: ...

class ThreadPoolStatus(_message.Message):
    __slots__ = ("overall", "name", "usage")
    OVERALL_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    USAGE_FIELD_NUMBER: _ClassVar[int]
    overall: StatusFlag.Status
    name: str
    usage: float
    def __init__(self, overall: _Optional[_Union[StatusFlag.Status, str]] = ..., name: _Optional[str] = ..., usage: _Optional[float] = ...) -> None: ...

class LoadAverageStatus(_message.Message):
    __slots__ = ("overall", "load", "cores")
    OVERALL_FIELD_NUMBER: _ClassVar[int]
    LOAD_FIELD_NUMBER: _ClassVar[int]
    CORES_FIELD_NUMBER: _ClassVar[int]
    overall: StatusFlag.Status
    load: float
    cores: int
    def __init__(self, overall: _Optional[_Union[StatusFlag.Status, str]] = ..., load: _Optional[float] = ..., cores: _Optional[int] = ...) -> None: ...

class ComputeNodeStatus(_message.Message):
    __slots__ = ("id", "overall", "tablets", "pools", "load")
    ID_FIELD_NUMBER: _ClassVar[int]
    OVERALL_FIELD_NUMBER: _ClassVar[int]
    TABLETS_FIELD_NUMBER: _ClassVar[int]
    POOLS_FIELD_NUMBER: _ClassVar[int]
    LOAD_FIELD_NUMBER: _ClassVar[int]
    id: str
    overall: StatusFlag.Status
    tablets: _containers.RepeatedCompositeFieldContainer[ComputeTabletStatus]
    pools: _containers.RepeatedCompositeFieldContainer[ThreadPoolStatus]
    load: LoadAverageStatus
    def __init__(self, id: _Optional[str] = ..., overall: _Optional[_Union[StatusFlag.Status, str]] = ..., tablets: _Optional[_Iterable[_Union[ComputeTabletStatus, _Mapping]]] = ..., pools: _Optional[_Iterable[_Union[ThreadPoolStatus, _Mapping]]] = ..., load: _Optional[_Union[LoadAverageStatus, _Mapping]] = ...) -> None: ...

class ComputeStatus(_message.Message):
    __slots__ = ("overall", "nodes", "tablets")
    OVERALL_FIELD_NUMBER: _ClassVar[int]
    NODES_FIELD_NUMBER: _ClassVar[int]
    TABLETS_FIELD_NUMBER: _ClassVar[int]
    overall: StatusFlag.Status
    nodes: _containers.RepeatedCompositeFieldContainer[ComputeNodeStatus]
    tablets: _containers.RepeatedCompositeFieldContainer[ComputeTabletStatus]
    def __init__(self, overall: _Optional[_Union[StatusFlag.Status, str]] = ..., nodes: _Optional[_Iterable[_Union[ComputeNodeStatus, _Mapping]]] = ..., tablets: _Optional[_Iterable[_Union[ComputeTabletStatus, _Mapping]]] = ...) -> None: ...

class LocationNode(_message.Message):
    __slots__ = ("id", "host", "port")
    ID_FIELD_NUMBER: _ClassVar[int]
    HOST_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    id: int
    host: str
    port: int
    def __init__(self, id: _Optional[int] = ..., host: _Optional[str] = ..., port: _Optional[int] = ...) -> None: ...

class LocationStoragePDisk(_message.Message):
    __slots__ = ("id", "path")
    ID_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    id: str
    path: str
    def __init__(self, id: _Optional[str] = ..., path: _Optional[str] = ...) -> None: ...

class LocationStorageVDisk(_message.Message):
    __slots__ = ("id", "pdisk")
    ID_FIELD_NUMBER: _ClassVar[int]
    PDISK_FIELD_NUMBER: _ClassVar[int]
    id: _containers.RepeatedScalarFieldContainer[str]
    pdisk: _containers.RepeatedCompositeFieldContainer[LocationStoragePDisk]
    def __init__(self, id: _Optional[_Iterable[str]] = ..., pdisk: _Optional[_Iterable[_Union[LocationStoragePDisk, _Mapping]]] = ...) -> None: ...

class LocationStorageGroup(_message.Message):
    __slots__ = ("id", "vdisk")
    ID_FIELD_NUMBER: _ClassVar[int]
    VDISK_FIELD_NUMBER: _ClassVar[int]
    id: _containers.RepeatedScalarFieldContainer[str]
    vdisk: LocationStorageVDisk
    def __init__(self, id: _Optional[_Iterable[str]] = ..., vdisk: _Optional[_Union[LocationStorageVDisk, _Mapping]] = ...) -> None: ...

class LocationStoragePool(_message.Message):
    __slots__ = ("name", "group")
    NAME_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    name: str
    group: LocationStorageGroup
    def __init__(self, name: _Optional[str] = ..., group: _Optional[_Union[LocationStorageGroup, _Mapping]] = ...) -> None: ...

class LocationStorage(_message.Message):
    __slots__ = ("node", "pool")
    NODE_FIELD_NUMBER: _ClassVar[int]
    POOL_FIELD_NUMBER: _ClassVar[int]
    node: LocationNode
    pool: LocationStoragePool
    def __init__(self, node: _Optional[_Union[LocationNode, _Mapping]] = ..., pool: _Optional[_Union[LocationStoragePool, _Mapping]] = ...) -> None: ...

class LocationComputePool(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class LocationComputeTablet(_message.Message):
    __slots__ = ("type", "id", "count")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    type: str
    id: _containers.RepeatedScalarFieldContainer[str]
    count: int
    def __init__(self, type: _Optional[str] = ..., id: _Optional[_Iterable[str]] = ..., count: _Optional[int] = ...) -> None: ...

class LocationCompute(_message.Message):
    __slots__ = ("node", "pool", "tablet")
    NODE_FIELD_NUMBER: _ClassVar[int]
    POOL_FIELD_NUMBER: _ClassVar[int]
    TABLET_FIELD_NUMBER: _ClassVar[int]
    node: LocationNode
    pool: LocationComputePool
    tablet: LocationComputeTablet
    def __init__(self, node: _Optional[_Union[LocationNode, _Mapping]] = ..., pool: _Optional[_Union[LocationComputePool, _Mapping]] = ..., tablet: _Optional[_Union[LocationComputeTablet, _Mapping]] = ...) -> None: ...

class LocationDatabase(_message.Message):
    __slots__ = ("name",)
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class Location(_message.Message):
    __slots__ = ("storage", "compute", "database")
    STORAGE_FIELD_NUMBER: _ClassVar[int]
    COMPUTE_FIELD_NUMBER: _ClassVar[int]
    DATABASE_FIELD_NUMBER: _ClassVar[int]
    storage: LocationStorage
    compute: LocationCompute
    database: LocationDatabase
    def __init__(self, storage: _Optional[_Union[LocationStorage, _Mapping]] = ..., compute: _Optional[_Union[LocationCompute, _Mapping]] = ..., database: _Optional[_Union[LocationDatabase, _Mapping]] = ...) -> None: ...

class IssueLog(_message.Message):
    __slots__ = ("id", "status", "message", "location", "reason", "type", "level", "listed", "count")
    ID_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    LOCATION_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    LISTED_FIELD_NUMBER: _ClassVar[int]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    id: str
    status: StatusFlag.Status
    message: str
    location: Location
    reason: _containers.RepeatedScalarFieldContainer[str]
    type: str
    level: int
    listed: int
    count: int
    def __init__(self, id: _Optional[str] = ..., status: _Optional[_Union[StatusFlag.Status, str]] = ..., message: _Optional[str] = ..., location: _Optional[_Union[Location, _Mapping]] = ..., reason: _Optional[_Iterable[str]] = ..., type: _Optional[str] = ..., level: _Optional[int] = ..., listed: _Optional[int] = ..., count: _Optional[int] = ...) -> None: ...

class DatabaseStatus(_message.Message):
    __slots__ = ("name", "overall", "storage", "compute")
    NAME_FIELD_NUMBER: _ClassVar[int]
    OVERALL_FIELD_NUMBER: _ClassVar[int]
    STORAGE_FIELD_NUMBER: _ClassVar[int]
    COMPUTE_FIELD_NUMBER: _ClassVar[int]
    name: str
    overall: StatusFlag.Status
    storage: StorageStatus
    compute: ComputeStatus
    def __init__(self, name: _Optional[str] = ..., overall: _Optional[_Union[StatusFlag.Status, str]] = ..., storage: _Optional[_Union[StorageStatus, _Mapping]] = ..., compute: _Optional[_Union[ComputeStatus, _Mapping]] = ...) -> None: ...

class SelfCheckResult(_message.Message):
    __slots__ = ("self_check_result", "issue_log", "database_status")
    SELF_CHECK_RESULT_FIELD_NUMBER: _ClassVar[int]
    ISSUE_LOG_FIELD_NUMBER: _ClassVar[int]
    DATABASE_STATUS_FIELD_NUMBER: _ClassVar[int]
    self_check_result: SelfCheck.Result
    issue_log: _containers.RepeatedCompositeFieldContainer[IssueLog]
    database_status: _containers.RepeatedCompositeFieldContainer[DatabaseStatus]
    def __init__(self, self_check_result: _Optional[_Union[SelfCheck.Result, str]] = ..., issue_log: _Optional[_Iterable[_Union[IssueLog, _Mapping]]] = ..., database_status: _Optional[_Iterable[_Union[DatabaseStatus, _Mapping]]] = ...) -> None: ...



# -*- coding: utf-8 -*-
from ydb import credentials, tracing
import grpc
import time
import abc
from datetime import datetime, timezone
import json
import os

try:
    import jwt
except ImportError:
    jwt = None

try:
    from yandex.cloud.iam.v1 import iam_token_service_pb2_grpc
    from yandex.cloud.iam.v1 import iam_token_service_pb2
except ImportError:
    try:
        # This attempt is to enable the IAM auth inside the YDB repository on GitHub
        from ydb.public.api.client.yc_public.iam import iam_token_service_pb2_grpc
        from ydb.public.api.client.yc_public.iam import iam_token_service_pb2
    except ImportError:
        try:
            # This attempt is to enable the IAM auth inside the YDB repository on Arcadia
            from contrib.ydb.public.api.client.yc_public.iam import iam_token_service_pb2_grpc
            from contrib.ydb.public.api.client.yc_public.iam import iam_token_service_pb2
        except ImportError:
            iam_token_service_pb2_grpc = None
            iam_token_service_pb2 = None

try:
    import requests
except ImportError:
    requests = None


DEFAULT_METADATA_URL = "http://169.254.169.254/computeMetadata/v1/instance/service-accounts/default/token"
YANDEX_CLOUD_IAM_TOKEN_SERVICE_URL = "https://iam.api.cloud.yandex.net/iam/v1/tokens"
YANDEX_CLOUD_JWT_ALGORITHM = "PS256"


def get_jwt(account_id, access_key_id, private_key, jwt_expiration_timeout, algorithm, token_service_url, subject=None):
    assert jwt is not None, "Install pyjwt library to use jwt tokens"
    now = time.time()
    now_utc = datetime.fromtimestamp(now, timezone.utc)
    exp_utc = datetime.fromtimestamp(now + jwt_expiration_timeout, timezone.utc)
    payload = {
        "iss": account_id,
        "aud": token_service_url,
        "iat": now_utc,
        "exp": exp_utc,
    }
    if subject is not None:
        payload["sub"] = subject
    return jwt.encode(
        key=private_key,
        algorithm=algorithm,
        headers={"typ": "JWT", "alg": algorithm, "kid": access_key_id},
        payload=payload,
    )


class TokenServiceCredentials(credentials.AbstractExpiringTokenCredentials):
    def __init__(self, iam_endpoint=None, iam_channel_credentials=None, tracer=None):
        super(TokenServiceCredentials, self).__init__(tracer)
        assert iam_token_service_pb2_grpc is not None, 'run pip install "ydb[yc]" to use service account credentials'
        self._get_token_request_timeout = 10
        self._iam_token_service_pb2 = iam_token_service_pb2
        self._iam_token_service_pb2_grpc = iam_token_service_pb2_grpc
        self._iam_endpoint = "iam.api.cloud.yandex.net:443" if iam_endpoint is None else iam_endpoint
        self._iam_channel_credentials = {} if iam_channel_credentials is None else iam_channel_credentials

    def _channel_factory(self):
        return grpc.secure_channel(
            self._iam_endpoint,
            grpc.ssl_channel_credentials(**self._iam_channel_credentials),
        )

    @abc.abstractmethod
    def _get_token_request(self):
        pass

    @tracing.with_trace()
    def _make_token_request(self):
        with self._channel_factory() as channel:
            tracing.trace(self.tracer, {"iam_token.from_service": True})
            stub = self._iam_token_service_pb2_grpc.IamTokenServiceStub(channel)
            response = stub.Create(self._get_token_request(), timeout=self._get_token_request_timeout)
            expires_in = max(0, response.expires_at.seconds - int(time.time()))
            return {"access_token": response.iam_token, "expires_in": expires_in}


class BaseJWTCredentials(abc.ABC):
    def __init__(self, account_id, access_key_id, private_key, algorithm, token_service_url, subject=None):
        self._account_id = account_id
        self._jwt_expiration_timeout = 60.0 * 60
        self._token_expiration_timeout = 120
        self._access_key_id = access_key_id
        self._private_key = private_key
        self._algorithm = algorithm
        self._token_service_url = token_service_url
        self._subject = subject

    def set_token_expiration_timeout(self, value):
        self._token_expiration_timeout = value
        return self

    @classmethod
    def from_file(cls, key_file, iam_endpoint=None, iam_channel_credentials=None):
        with open(os.path.expanduser(key_file), "r") as r:
            key = r.read()

        return cls.from_content(key, iam_endpoint=iam_endpoint, iam_channel_credentials=iam_channel_credentials)

    @classmethod
    def from_content(cls, key, iam_endpoint=None, iam_channel_credentials=None):
        key_json = json.loads(key)
        account_id = key_json.get("service_account_id", None)
        if account_id is None:
            account_id = key_json.get("user_account_id", None)
        return cls(
            account_id,
            key_json["id"],
            key_json["private_key"],
            iam_endpoint=iam_endpoint,
            iam_channel_credentials=iam_channel_credentials,
        )

    def _get_jwt(self):
        return get_jwt(
            self._account_id,
            self._access_key_id,
            self._private_key,
            self._jwt_expiration_timeout,
            self._algorithm,
            self._token_service_url,
            self._subject,
        )


class JWTIamCredentials(TokenServiceCredentials, BaseJWTCredentials):
    def __init__(
        self,
        account_id,
        access_key_id,
        private_key,
        iam_endpoint=None,
        iam_channel_credentials=None,
    ):
        TokenServiceCredentials.__init__(self, iam_endpoint, iam_channel_credentials)
        BaseJWTCredentials.__init__(
            self, account_id, access_key_id, private_key, YANDEX_CLOUD_JWT_ALGORITHM, YANDEX_CLOUD_IAM_TOKEN_SERVICE_URL
        )

    def _get_token_request(self):
        return self._iam_token_service_pb2.CreateIamTokenRequest(jwt=self._get_jwt())


class YandexPassportOAuthIamCredentials(TokenServiceCredentials):
    def __init__(
        self,
        yandex_passport_oauth_token,
        iam_endpoint=None,
        iam_channel_credentials=None,
    ):
        self._yandex_passport_oauth_token = yandex_passport_oauth_token
        super(YandexPassportOAuthIamCredentials, self).__init__(iam_endpoint, iam_channel_credentials)

    def _get_token_request(self):
        return iam_token_service_pb2.CreateIamTokenRequest(
            yandex_passport_oauth_token=self._yandex_passport_oauth_token
        )


class MetadataUrlCredentials(credentials.AbstractExpiringTokenCredentials):
    def __init__(self, metadata_url=None, tracer=None):
        """
        :param metadata_url: Metadata url
        :param ydb.Tracer tracer: ydb tracer
        """
        super(MetadataUrlCredentials, self).__init__(tracer)
        assert requests is not None, "Install requests library to use metadata credentials provider"
        self.extra_error_message = (
            "Check that metadata service configured properly since we failed to fetch it from metadata_url."
        )
        self._metadata_url = DEFAULT_METADATA_URL if metadata_url is None else metadata_url

    @tracing.with_trace()
    def _make_token_request(self):
        response = requests.get(self._metadata_url, headers={"Metadata-Flavor": "Google"}, timeout=3)
        response.raise_for_status()
        return json.loads(response.text)


class ServiceAccountCredentials(JWTIamCredentials):
    def __init__(
        self,
        service_account_id,
        access_key_id,
        private_key,
        iam_endpoint=None,
        iam_channel_credentials=None,
    ):
        super(ServiceAccountCredentials, self).__init__(
            service_account_id,
            access_key_id,
            private_key,
            iam_endpoint,
            iam_channel_credentials,
        )



import asyncio
import copy
import functools
from collections import deque
from typing import List

import pytest

from ydb._topic_common.test_helpers import wait_condition
from ydb._topic_reader import topic_reader_asyncio
from ydb._topic_reader.datatypes import PartitionSession


@pytest.mark.asyncio
class TestPartitionSession:
    session_comitted_offset = 10

    @pytest.fixture
    def session(self) -> PartitionSession:
        return PartitionSession(
            id=1,
            state=PartitionSession.State.Active,
            topic_path="",
            partition_id=1,
            committed_offset=self.session_comitted_offset,
            reader_reconnector_id=1,
            reader_stream_id=1,
        )

    @pytest.mark.parametrize(
        "offsets_waited,notify_offset,offsets_notified,offsets_waited_rest",
        [
            ([1], 1, [1], []),
            ([1], 10, [1], []),
            ([1, 2, 3], 10, [1, 2, 3], []),
            ([1, 2, 10, 20], 10, [1, 2, 10], [20]),
            ([10, 20], 1, [], [10, 20]),
        ],
    )
    async def test_ack_notify(
        self,
        session,
        offsets_waited: List[int],
        notify_offset: int,
        offsets_notified: List[int],
        offsets_waited_rest: List[int],
    ):
        notified = []

        for offset in offsets_waited:
            fut = asyncio.Future()

            def add_notify(future, notified_offset):
                notified.append(notified_offset)

            fut.add_done_callback(functools.partial(add_notify, notified_offset=offset))
            waiter = PartitionSession.CommitAckWaiter(offset, fut)
            session._ack_waiters.append(waiter)

        session.ack_notify(notify_offset)
        assert session._ack_waiters == deque(
            [PartitionSession.CommitAckWaiter(offset, asyncio.Future()) for offset in offsets_waited_rest]
        )

        await wait_condition(lambda: len(notified) == len(offsets_notified))

        notified.sort()
        assert notified == offsets_notified
        assert session.committed_offset == notify_offset

    # noinspection PyTypeChecker
    @pytest.mark.parametrize(
        "original,add,is_done,result",
        [
            (
                [],
                session_comitted_offset - 5,
                True,
                [],
            ),
            (
                [PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None)],
                session_comitted_offset + 0,
                True,
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                ],
            ),
            (
                [],
                session_comitted_offset + 5,
                False,
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                ],
            ),
            (
                [PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None)],
                session_comitted_offset + 6,
                False,
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 6, None),
                ],
            ),
            (
                [PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None)],
                session_comitted_offset + 4,
                False,
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 4, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                ],
            ),
            (
                [PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None)],
                session_comitted_offset + 100,
                False,
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 100, None),
                ],
            ),
            (
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 100, None),
                ],
                session_comitted_offset + 50,
                False,
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 50, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 100, None),
                ],
            ),
            (
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 7, None),
                ],
                session_comitted_offset + 6,
                False,
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 6, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 7, None),
                ],
            ),
            (
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 100, None),
                ],
                session_comitted_offset + 6,
                False,
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 6, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 100, None),
                ],
            ),
            (
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 100, None),
                ],
                session_comitted_offset + 99,
                False,
                [
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 5, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 99, None),
                    PartitionSession.CommitAckWaiter(session_comitted_offset + 100, None),
                ],
            ),
        ],
    )
    async def test_add_waiter(
        self,
        session,
        original: List[PartitionSession.CommitAckWaiter],
        add: int,
        is_done: bool,
        result: List[PartitionSession.CommitAckWaiter],
    ):
        session._ack_waiters = copy.deepcopy(original)
        res = session.add_waiter(add)
        assert result == session._ack_waiters
        assert res.future.done() == is_done

    async def test_close_notify_waiters(self, session):
        waiter = session.add_waiter(session.committed_offset + 1)
        session.close()

        with pytest.raises(topic_reader_asyncio.PublicTopicReaderPartitionExpiredError):
            waiter.future.result()

    async def test_close_twice(self, session):
        session.close()
        session.close()

    async def test_commit_after_close(self, session):
        session.close()

        with pytest.raises(topic_reader_asyncio.PublicTopicReaderPartitionExpiredError):
            session.add_waiter(session.committed_offset + 1)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_export_v1.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v5.protos import ydb_export_pb2 as protos_dot_ydb__export__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13ydb_export_v1.proto\x12\rYdb.Export.V1\x1a\x17protos/ydb_export.proto2\xf6\x01\n\rExportService\x12K\n\nExportToYt\x12\x1d.Ydb.Export.ExportToYtRequest\x1a\x1e.Ydb.Export.ExportToYtResponse\x12K\n\nExportToS3\x12\x1d.Ydb.Export.ExportToS3Request\x1a\x1e.Ydb.Export.ExportToS3Response\x12K\n\nExportToFs\x12\x1d.Ydb.Export.ExportToFsRequest\x1a\x1e.Ydb.Export.ExportToFsResponseBQ\n\x18tech.ydb.proto.export.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Export_V1b\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ydb_export_v1_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030tech.ydb.proto.export.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Export_V1'
  _EXPORTSERVICE._serialized_start=64
  _EXPORTSERVICE._serialized_end=310
# @@protoc_insertion_point(module_scope)



from . import issues, _session_impl, _apis, types, convert
import functools


def reset_tx_id_handler(func):
    @functools.wraps(func)
    def decorator(rpc_state, response_pb, session_state, tx_state, *args, **kwargs):
        try:
            return func(rpc_state, response_pb, session_state, tx_state, *args, **kwargs)
        except issues.Error:
            tx_state.tx_id = None
            tx_state.dead = True
            raise

    return decorator


def not_found_handler(func):
    @functools.wraps(func)
    def decorator(rpc_state, response_pb, session_state, tx_state, query, *args, **kwargs):
        try:
            return func(rpc_state, response_pb, session_state, tx_state, query, *args, **kwargs)
        except issues.NotFound:
            session_state.erase(query)
            raise

    return decorator


def wrap_tx_factory_handler(func):
    @functools.wraps(func)
    def decorator(session_state, tx_state, *args, **kwargs):
        if tx_state.dead:
            raise issues.PreconditionFailed("Failed to perform action on broken transaction context!")
        return func(session_state, tx_state, *args, **kwargs)

    return decorator


@_session_impl.bad_session_handler
@reset_tx_id_handler
def wrap_result_on_rollback_or_commit_tx(rpc_state, response_pb, session_state, tx_state, tx):
    session_state.complete_query()
    issues._process_response(response_pb.operation)
    # transaction successfully committed or rolled back
    tx_state.tx_id = None
    return tx


@_session_impl.bad_session_handler
def wrap_tx_begin_response(rpc_state, response_pb, session_state, tx_state, tx):
    session_state.complete_query()
    issues._process_response(response_pb.operation)
    message = _apis.ydb_table.BeginTransactionResult()
    response_pb.operation.result.Unpack(message)
    tx_state.tx_id = message.tx_meta.id
    return tx


@wrap_tx_factory_handler
def begin_request_factory(session_state, tx_state):
    request = _apis.ydb_table.BeginTransactionRequest()
    request = session_state.start_query().attach_request(request)
    request.tx_settings.MergeFrom(_construct_tx_settings(tx_state))
    return request


@wrap_tx_factory_handler
def rollback_request_factory(session_state, tx_state):
    request = _apis.ydb_table.RollbackTransactionRequest()
    request.tx_id = tx_state.tx_id
    request = session_state.start_query().attach_request(request)
    return request


@wrap_tx_factory_handler
def commit_request_factory(session_state, tx_state):
    """
    Constructs commit request
    """
    request = _apis.ydb_table.CommitTransactionRequest()
    request.tx_id = tx_state.tx_id
    request = session_state.start_query().attach_request(request)
    return request


class TxState(object):
    __slots__ = ("tx_id", "tx_mode", "dead", "initialized")

    def __init__(self, tx_mode):
        """
        Holds transaction context manager info
        :param tx_mode: A mode of transaction
        """
        self.tx_id = None
        self.tx_mode = tx_mode
        self.dead = False
        self.initialized = False


def _construct_tx_settings(tx_state):
    tx_settings = _apis.ydb_table.TransactionSettings()
    mode_property = getattr(tx_settings, tx_state.tx_mode.name)
    mode_property.MergeFrom(tx_state.tx_mode.settings)
    return tx_settings


@wrap_tx_factory_handler
def execute_request_factory(session_state, tx_state, query, parameters, commit_tx, settings):
    data_query, query_id = session_state.lookup(query)
    parameters_types = {}

    is_data_query = False

    if query_id is not None:
        query_pb = _apis.ydb_table.Query(id=query_id)
        parameters_types = data_query.parameters_types
    else:
        if data_query is not None:
            # client cache disabled for send query text every time
            yql_text = data_query.yql_text
            parameters_types = data_query.parameters_types
            is_data_query = True
        elif isinstance(query, types.DataQuery):
            yql_text = query.yql_text
            parameters_types = query.parameters_types
            is_data_query = True
        else:
            yql_text = query
        query_pb = _apis.ydb_table.Query(yql_text=yql_text)
    request = _apis.ydb_table.ExecuteDataQueryRequest(parameters=convert.parameters_to_pb(parameters_types, parameters))

    if query_id is not None:
        # SDK not send query text and nothing save to cache
        keep_in_cache = False
    elif settings is not None and hasattr(settings, "keep_in_cache"):
        keep_in_cache = settings.keep_in_cache
    elif parameters:
        keep_in_cache = True
    elif is_data_query:
        keep_in_cache = True
    else:
        keep_in_cache = False

    if keep_in_cache:
        request.query_cache_policy.keep_in_cache = True

    request.query.MergeFrom(query_pb)
    tx_control = _apis.ydb_table.TransactionControl()
    tx_control.commit_tx = commit_tx
    if tx_state.tx_id is not None:
        tx_control.tx_id = tx_state.tx_id
    else:
        tx_control.begin_tx.MergeFrom(_construct_tx_settings(tx_state))
    request.tx_control.MergeFrom(tx_control)
    request = session_state.start_query().attach_request(request)
    return request


@_session_impl.bad_session_handler
@reset_tx_id_handler
@not_found_handler
def wrap_result_and_tx_id(rpc_state, response_pb, session_state, tx_state, query):
    session_state.complete_query()
    issues._process_response(response_pb.operation)
    message = _apis.ydb_table.ExecuteQueryResult()
    response_pb.operation.result.Unpack(message)
    if message.query_meta.id and isinstance(query, types.DataQuery):
        session_state.keep(query, message.query_meta.id)
    tx_state.tx_id = None if not message.tx_meta.id else message.tx_meta.id
    return convert.ResultSets(message.result_sets, session_state.table_client_settings)



from __future__ import annotations

import asyncio
import concurrent.futures
import gzip
import typing
from asyncio import Task
from collections import defaultdict, OrderedDict
from typing import Optional, Set, Dict, Union, Callable

import ydb
from .. import _apis, issues
from .._topic_common import common as topic_common
from .._utilities import AtomicCounter
from ..aio import Driver
from ..issues import Error as YdbError, _process_response
from . import datatypes
from . import events
from . import topic_reader
from .._grpc.grpcwrapper.common_utils import (
    IGrpcWrapperAsyncIO,
    SupportedDriverType,
    to_thread,
    GrpcWrapperAsyncIO,
)
from .._grpc.grpcwrapper.ydb_topic import (
    StreamReadMessage,
    UpdateTokenRequest,
    UpdateTokenResponse,
    UpdateOffsetsInTransactionRequest,
    Codec,
)
from .._errors import check_retriable_error
import logging

from ..query.base import TxEvent

if typing.TYPE_CHECKING:
    from ..query.transaction import BaseQueryTxContext

from .._constants import DEFAULT_INITIAL_RESPONSE_TIMEOUT

logger = logging.getLogger(__name__)


class TopicReaderError(YdbError):
    pass


class PublicTopicReaderUnexpectedCodecError(YdbError):
    pass


class PublicTopicReaderPartitionExpiredError(TopicReaderError):
    """
    Commit message when partition read session are dropped.
    It is ok - the message/batch will not commit to server and will receive in other read session
    (with this or other reader).
    """

    def __init__(self, message: str = "Topic reader partition session is closed"):
        super().__init__(message)


class TopicReaderStreamClosedError(TopicReaderError):
    def __init__(self):
        super().__init__("Topic reader stream is closed")


class TopicReaderClosedError(TopicReaderError):
    def __init__(self):
        super().__init__("Topic reader is closed already")


class PublicAsyncIOReader:
    _loop: asyncio.AbstractEventLoop
    _closed: bool
    _settings: topic_reader.PublicReaderSettings
    _reconnector: ReaderReconnector
    _parent: typing.Any  # need for prevent close parent client by GC

    def __init__(
        self,
        driver: Driver,
        settings: topic_reader.PublicReaderSettings,
        *,
        _parent=None,
    ):
        self._loop = asyncio.get_running_loop()
        self._closed = False
        self._settings = settings
        self._reconnector = ReaderReconnector(driver, settings, self._loop)
        self._parent = _parent

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def __del__(self):
        if not self._closed:
            try:
                logger.debug("Topic reader was not closed properly. Consider using method close().")
                task = self._loop.create_task(self.close(flush=False))
                topic_common.wrap_set_name_for_asyncio_task(task, task_name="close reader")
            except BaseException:
                logger.warning("Something went wrong during reader close in __del__")

    async def wait_message(self):
        """
        Wait at least one message from reader.
        """
        await self._reconnector.wait_message()

    async def receive_batch(
        self,
        max_messages: typing.Union[int, None] = None,
    ) -> typing.Union[datatypes.PublicBatch, None]:
        """
        Get one messages batch from reader.
        All messages in a batch from same partition.

        use asyncio.wait_for for wait with timeout.
        """
        logger.debug("receive_batch max_messages=%s", max_messages)
        await self._reconnector.wait_message()
        return self._reconnector.receive_batch_nowait(
            max_messages=max_messages,
        )

    async def receive_batch_with_tx(
        self,
        tx: "BaseQueryTxContext",
        max_messages: typing.Union[int, None] = None,
    ) -> typing.Union[datatypes.PublicBatch, None]:
        """
        Get one messages batch with tx from reader.
        All messages in a batch from same partition.

        use asyncio.wait_for for wait with timeout.
        """
        logger.debug("receive_batch_with_tx tx=%s max_messages=%s", tx, max_messages)
        await self._reconnector.wait_message()
        return self._reconnector.receive_batch_with_tx_nowait(
            tx=tx,
            max_messages=max_messages,
        )

    async def receive_message(self) -> typing.Optional[datatypes.PublicMessage]:
        """
        Block until receive new message

        use asyncio.wait_for for wait with timeout.
        """
        logger.debug("receive_message")
        await self._reconnector.wait_message()
        return self._reconnector.receive_message_nowait()

    def commit(self, batch: typing.Union[datatypes.PublicMessage, datatypes.PublicBatch]):
        """
        Write commit message to a buffer.

        For the method no way check the commit result
        (for example if lost connection - commits will not re-send and committed messages will receive again).
        """
        logger.debug("commit message or batch")
        if self._settings.consumer is None:
            raise issues.Error("Commit operations are not supported for topic reader without consumer.")

        try:
            self._reconnector.commit(batch)
        except PublicTopicReaderPartitionExpiredError:
            pass

    async def commit_with_ack(self, batch: typing.Union[datatypes.PublicMessage, datatypes.PublicBatch]):
        """
        write commit message to a buffer and wait ack from the server.

        use asyncio.wait_for for wait with timeout.

        may raise ydb.TopicReaderPartitionExpiredError, the error mean reader partition closed from server
        before receive commit ack. Message may be acked or not (if not - it will send in other read session,
        to this or other reader).
        """
        logger.debug("commit_with_ack message or batch")
        if self._settings.consumer is None:
            raise issues.Error("Commit operations are not supported for topic reader without consumer.")

        waiter = self._reconnector.commit(batch)
        await waiter.future

    async def close(self, flush: bool = True):
        if self._closed:
            raise TopicReaderClosedError()

        logger.debug("Close topic reader")
        self._closed = True
        await self._reconnector.close(flush)
        logger.debug("Topic reader was closed")

    @property
    def read_session_id(self) -> Optional[str]:
        return self._reconnector.read_session_id


class ReaderReconnector:
    _static_reader_reconnector_counter = AtomicCounter()

    _id: int
    _settings: topic_reader.PublicReaderSettings
    _driver: Driver
    _background_tasks: Set[Task]

    _state_changed: asyncio.Event
    _stream_reader: Optional["ReaderStream"]
    _first_error: asyncio.Future[YdbError]
    _tx_to_batches_map: Dict[str, typing.List[datatypes.PublicBatch]]

    def __init__(
        self,
        driver: Driver,
        settings: topic_reader.PublicReaderSettings,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self._id = ReaderReconnector._static_reader_reconnector_counter.inc_and_get()
        self._settings = settings
        self._driver = driver
        self._loop = loop if loop is not None else asyncio.get_running_loop()
        self._background_tasks = set()
        logger.debug("init reader reconnector id=%s", self._id)

        self._state_changed = asyncio.Event()
        self._stream_reader = None
        self._background_tasks.add(asyncio.create_task(self._connection_loop()))
        self._first_error = asyncio.get_running_loop().create_future()

        self._tx_to_batches_map = dict()

    async def _connection_loop(self):
        attempt = 0
        while True:
            try:
                logger.debug("reader %s connect attempt %s", self._id, attempt)
                self._stream_reader = await ReaderStream.create(self._id, self._driver, self._settings)
                logger.debug("reader %s connected stream %s", self._id, self._stream_reader._id)
                attempt = 0
                self._state_changed.set()
                await self._stream_reader.wait_error()
            except BaseException as err:
                logger.debug("reader %s, attempt %s connection loop error %s", self._id, attempt, err)
                retry_info = check_retriable_error(err, self._settings._retry_settings(), attempt)
                if not retry_info.is_retriable:
                    logger.debug("reader %s stop connection loop due to %s", self._id, err)
                    self._set_first_error(err)
                    return

                logger.debug("sleep before retry for %s seconds", retry_info.sleep_timeout_seconds)

                await asyncio.sleep(retry_info.sleep_timeout_seconds)

                attempt += 1
            finally:
                if self._stream_reader is not None:
                    # noinspection PyBroadException
                    try:
                        await self._stream_reader.close(flush=False)
                    except BaseException:
                        # supress any error on close stream reader
                        pass

    async def wait_message(self):
        while True:
            if self._first_error.done():
                raise self._first_error.result()

            if self._stream_reader:
                try:
                    await self._stream_reader.wait_messages()
                    return
                except YdbError:
                    pass  # handle errors in reconnection loop

            await self._state_changed.wait()
            self._state_changed.clear()

    def receive_batch_nowait(self, max_messages: Optional[int] = None):
        return self._stream_reader.receive_batch_nowait(
            max_messages=max_messages,
        )

    def receive_batch_with_tx_nowait(self, tx: "BaseQueryTxContext", max_messages: Optional[int] = None):
        batch = self._stream_reader.receive_batch_nowait(
            max_messages=max_messages,
        )

        self._init_tx(tx)

        self._tx_to_batches_map[tx.tx_id].append(batch)

        tx._add_callback(TxEvent.AFTER_COMMIT, batch._update_partition_offsets, self._loop)

        return batch

    def receive_message_nowait(self):
        return self._stream_reader.receive_message_nowait()

    def _init_tx(self, tx: "BaseQueryTxContext"):
        if tx.tx_id not in self._tx_to_batches_map:  # Init tx callbacks
            self._tx_to_batches_map[tx.tx_id] = []
            tx._add_callback(TxEvent.BEFORE_COMMIT, self._commit_batches_with_tx, self._loop)
            tx._add_callback(TxEvent.AFTER_COMMIT, self._handle_after_tx_commit, self._loop)
            tx._add_callback(TxEvent.AFTER_ROLLBACK, self._handle_after_tx_rollback, self._loop)

    async def _commit_batches_with_tx(self, tx: "BaseQueryTxContext"):
        grouped_batches = defaultdict(lambda: defaultdict(list))
        for batch in self._tx_to_batches_map[tx.tx_id]:
            grouped_batches[batch._partition_session.topic_path][batch._partition_session.partition_id].append(batch)

        request = UpdateOffsetsInTransactionRequest(tx=tx._tx_identity(), consumer=self._settings.consumer, topics=[])

        for topic_path in grouped_batches:
            topic_offsets = UpdateOffsetsInTransactionRequest.TopicOffsets(path=topic_path, partitions=[])
            for partition_id in grouped_batches[topic_path]:
                partition_offsets = UpdateOffsetsInTransactionRequest.TopicOffsets.PartitionOffsets(
                    partition_id=partition_id,
                    partition_offsets=[
                        batch._commit_get_offsets_range() for batch in grouped_batches[topic_path][partition_id]
                    ],
                )
                topic_offsets.partitions.append(partition_offsets)
            request.topics.append(topic_offsets)

        try:
            return await self._do_commit_batches_with_tx_call(request)
        except BaseException:
            exc = issues.ClientInternalError("Failed to update offsets in tx.")
            tx._set_external_error(exc)
            self._stream_reader._set_first_error(exc)
        finally:
            del self._tx_to_batches_map[tx.tx_id]

    async def _do_commit_batches_with_tx_call(self, request: UpdateOffsetsInTransactionRequest):
        args = [
            request.to_proto(),
            _apis.TopicService.Stub,
            _apis.TopicService.UpdateOffsetsInTransaction,
            topic_common.wrap_operation,
        ]

        if asyncio.iscoroutinefunction(self._driver.__call__):
            res = await self._driver(*args)
        else:
            res = await to_thread(self._driver, *args, executor=None)

        return res

    async def _handle_after_tx_rollback(self, tx: "BaseQueryTxContext", exc: Optional[BaseException]) -> None:
        if tx.tx_id in self._tx_to_batches_map:
            del self._tx_to_batches_map[tx.tx_id]
        exc = issues.ClientInternalError("Reconnect due to transaction rollback")
        self._stream_reader._set_first_error(exc)

    async def _handle_after_tx_commit(self, tx: "BaseQueryTxContext", exc: Optional[BaseException]) -> None:
        if tx.tx_id in self._tx_to_batches_map:
            del self._tx_to_batches_map[tx.tx_id]

        if exc is not None:
            self._stream_reader._set_first_error(
                issues.ClientInternalError("Reconnect due to transaction commit failed")
            )

    def commit(self, batch: datatypes.ICommittable) -> datatypes.PartitionSession.CommitAckWaiter:
        return self._stream_reader.commit(batch)

    async def close(self, flush: bool):
        logger.debug("reader reconnector %s close", self._id)
        if self._stream_reader:
            await self._stream_reader.close(flush)
        for task in self._background_tasks:
            task.cancel()

        await asyncio.wait(self._background_tasks)

    async def flush(self):
        if self._stream_reader:
            await self._stream_reader.flush()

    def _set_first_error(self, err: issues.Error):
        try:
            self._first_error.set_result(err)
            self._state_changed.set()
        except asyncio.InvalidStateError:
            # skip if already has result
            pass

    @property
    def read_session_id(self) -> Optional[str]:
        if not self._stream_reader:
            return None
        return self._stream_reader._session_id


class ReaderStream:
    _static_id_counter = AtomicCounter()

    _loop: asyncio.AbstractEventLoop
    _id: int
    _reader_reconnector_id: int
    _session_id: str
    _stream: Optional[IGrpcWrapperAsyncIO]
    _started: bool
    _background_tasks: Set[asyncio.Task]
    _partition_sessions: Dict[int, datatypes.PartitionSession]
    _buffer_size_bytes: int  # use for init request, then for debug purposes only
    _decode_executor: concurrent.futures.Executor
    _decoders: Dict[int, typing.Callable[[bytes], bytes]]  # dict[codec_code] func(encoded_bytes)->decoded_bytes

    if typing.TYPE_CHECKING:
        _batches_to_decode: asyncio.Queue[datatypes.PublicBatch]
    else:
        _batches_to_decode: asyncio.Queue

    _state_changed: asyncio.Event
    _closed: bool
    _message_batches: typing.Dict[int, datatypes.PublicBatch]  # keys are partition session ID
    _first_error: asyncio.Future[YdbError]

    _update_token_interval: Union[int, float]
    _update_token_event: asyncio.Event
    _get_token_function: Callable[[], str]
    _settings: topic_reader.PublicReaderSettings

    def __init__(
        self,
        reader_reconnector_id: int,
        settings: topic_reader.PublicReaderSettings,
        get_token_function: Optional[Callable[[], str]] = None,
    ):
        self._loop = asyncio.get_running_loop()
        self._id = ReaderStream._static_id_counter.inc_and_get()
        self._reader_reconnector_id = reader_reconnector_id
        self._session_id = "not initialized"
        self._stream = None
        self._started = False
        self._background_tasks = set()
        self._partition_sessions = dict()
        self._buffer_size_bytes = settings.buffer_size_bytes
        self._decode_executor = settings.decoder_executor

        self._decoders = {Codec.CODEC_GZIP: gzip.decompress}
        if settings.decoders:
            self._decoders.update(settings.decoders)

        self._state_changed = asyncio.Event()
        self._closed = False
        self._first_error = asyncio.get_running_loop().create_future()
        self._batches_to_decode = asyncio.Queue()
        self._message_batches = OrderedDict()

        self._update_token_interval = settings.update_token_interval
        self._get_token_function = get_token_function
        self._update_token_event = asyncio.Event()

        self._settings = settings

        logger.debug("created ReaderStream id=%s reconnector=%s", self._id, self._reader_reconnector_id)

    @staticmethod
    async def create(
        reader_reconnector_id: int,
        driver: SupportedDriverType,
        settings: topic_reader.PublicReaderSettings,
    ) -> "ReaderStream":
        stream = GrpcWrapperAsyncIO(StreamReadMessage.FromServer.from_proto)

        await stream.start(driver, _apis.TopicService.Stub, _apis.TopicService.StreamRead)

        creds = driver._credentials
        reader = ReaderStream(
            reader_reconnector_id,
            settings,
            get_token_function=creds.get_auth_token if creds else None,
        )
        await reader._start(stream, settings._init_message())
        logger.debug("reader stream %s started session=%s", reader._id, reader._session_id)
        return reader

    async def _start(self, stream: IGrpcWrapperAsyncIO, init_message: StreamReadMessage.InitRequest):
        if self._started:
            raise TopicReaderError("Double start ReaderStream")

        self._started = True
        self._stream = stream
        logger.debug("reader stream %s send init request", self._id)

        stream.write(StreamReadMessage.FromClient(client_message=init_message))
        try:
            init_response = await stream.receive(
                timeout=DEFAULT_INITIAL_RESPONSE_TIMEOUT
            )  # type: StreamReadMessage.FromServer
        except asyncio.TimeoutError:
            raise TopicReaderError("Timeout waiting for init response")

        if isinstance(init_response.server_message, StreamReadMessage.InitResponse):
            self._session_id = init_response.server_message.session_id
            logger.debug("reader stream %s initialized session=%s", self._id, self._session_id)
        else:
            raise TopicReaderError("Unexpected message after InitRequest: %s", init_response)

        self._update_token_event.set()

        self._background_tasks.add(
            topic_common.wrap_set_name_for_asyncio_task(
                asyncio.create_task(self._read_messages_loop()),
                task_name="read_messages_loop",
            ),
        )
        self._background_tasks.add(
            topic_common.wrap_set_name_for_asyncio_task(
                asyncio.create_task(self._decode_batches_loop()),
                task_name="decode_batches",
            ),
        )
        if self._get_token_function:
            self._background_tasks.add(
                topic_common.wrap_set_name_for_asyncio_task(
                    asyncio.create_task(self._update_token_loop()),
                    task_name="update_token_loop",
                ),
            )
        self._background_tasks.add(
            topic_common.wrap_set_name_for_asyncio_task(
                asyncio.create_task(self._handle_background_errors()),
                task_name="handle_background_errors",
            ),
        )

    async def wait_error(self):
        raise await self._first_error

    async def wait_messages(self):
        while True:
            if self._get_first_error():
                raise self._get_first_error()

            if self._message_batches:
                return

            await self._state_changed.wait()
            self._state_changed.clear()

    def _get_first_batch(self) -> typing.Tuple[int, datatypes.PublicBatch]:
        partition_session_id, batch = self._message_batches.popitem(last=False)
        return partition_session_id, batch

    def _return_batch_to_queue(self, part_sess_id: int, batch: datatypes.PublicBatch):
        self._message_batches[part_sess_id] = batch

        # In case of auto-split we should return all parent messages ASAP
        # without queue rotation to prevent child's messages before parent's.
        if part_sess_id in self._partition_sessions and self._partition_sessions[part_sess_id].ended:
            self._message_batches.move_to_end(part_sess_id, last=False)

    def receive_batch_nowait(self, max_messages: Optional[int] = None):
        if self._get_first_error():
            raise self._get_first_error()

        if not self._message_batches:
            return None

        part_sess_id, batch = self._get_first_batch()

        if max_messages is None or len(batch.messages) <= max_messages:
            self._buffer_release_bytes(batch._bytes_size)
            return batch

        cutted_batch = batch._pop_batch(message_count=max_messages)

        self._return_batch_to_queue(part_sess_id, batch)

        self._buffer_release_bytes(cutted_batch._bytes_size)

        return cutted_batch

    def receive_message_nowait(self):
        if self._get_first_error():
            raise self._get_first_error()

        if not self._message_batches:
            return None

        part_sess_id, batch = self._get_first_batch()

        message, msgs_left = batch._pop()

        if not msgs_left:
            self._buffer_release_bytes(batch._bytes_size)
        else:
            # TODO: we should somehow release bytes from single message as well
            self._return_batch_to_queue(part_sess_id, batch)

        return message

    def commit(self, batch: datatypes.ICommittable) -> datatypes.PartitionSession.CommitAckWaiter:
        partition_session = batch._commit_get_partition_session()

        if partition_session.reader_reconnector_id != self._reader_reconnector_id:
            raise TopicReaderError("reader can commit only self-produced messages")

        if partition_session.reader_stream_id != self._id:
            raise PublicTopicReaderPartitionExpiredError("commit messages after reconnect to server")

        if partition_session.id not in self._partition_sessions:
            raise PublicTopicReaderPartitionExpiredError("commit messages after server stop the partition read session")

        commit_range = batch._commit_get_offsets_range()
        waiter = partition_session.add_waiter(commit_range.end)

        if not waiter.future.done():
            client_message = StreamReadMessage.CommitOffsetRequest(
                commit_offsets=[
                    StreamReadMessage.CommitOffsetRequest.PartitionCommitOffset(
                        partition_session_id=partition_session.id,
                        offsets=[commit_range],
                    )
                ]
            )
            self._stream.write(StreamReadMessage.FromClient(client_message=client_message))

        return waiter

    async def _handle_background_errors(self):
        done, _ = await asyncio.wait(self._background_tasks, return_when=asyncio.FIRST_EXCEPTION)
        for f in done:
            f = f  # type: asyncio.Future
            err = f.exception()
            if not isinstance(err, ydb.Error):
                old_err = err
                err = ydb.Error("Background process failed unexpected")
                err.__cause__ = old_err
            self._set_first_error(err)

    async def _read_messages_loop(self):
        try:
            logger.debug("reader stream %s start read loop", self._id)
            self._stream.write(
                StreamReadMessage.FromClient(
                    client_message=StreamReadMessage.ReadRequest(
                        bytes_size=self._buffer_size_bytes,
                    ),
                )
            )
            while True:
                try:
                    message = await self._stream.receive()  # type: StreamReadMessage.FromServer
                    _process_response(message.server_status)

                    if isinstance(message.server_message, StreamReadMessage.ReadResponse):
                        logger.debug("reader stream %s read %s bytes", self._id, message.server_message.bytes_size)
                        self._on_read_response(message.server_message)

                    elif isinstance(message.server_message, StreamReadMessage.CommitOffsetResponse):
                        self._on_commit_response(message.server_message)

                    elif isinstance(
                        message.server_message,
                        StreamReadMessage.StartPartitionSessionRequest,
                    ):
                        logger.debug(
                            "reader stream %s start partition %s",
                            self._id,
                            message.server_message.partition_session.partition_session_id,
                        )
                        await self._on_start_partition_session(message.server_message)

                    elif isinstance(
                        message.server_message,
                        StreamReadMessage.StopPartitionSessionRequest,
                    ):
                        logger.debug(
                            "reader stream %s stop partition %s",
                            self._id,
                            message.server_message.partition_session_id,
                        )
                        self._on_partition_session_stop(message.server_message)

                    elif isinstance(
                        message.server_message,
                        StreamReadMessage.EndPartitionSession,
                    ):
                        logger.debug(
                            "reader stream %s end partition %s",
                            self._id,
                            message.server_message.partition_session_id,
                        )
                        self._on_end_partition_session(message.server_message)

                    elif isinstance(message.server_message, UpdateTokenResponse):
                        self._update_token_event.set()

                    else:
                        raise issues.UnexpectedGrpcMessage(
                            "Unexpected message in _read_messages_loop: %s" % type(message.server_message)
                        )
                except issues.UnexpectedGrpcMessage as e:
                    logger.exception("unexpected message in stream reader: %s" % e)

                self._state_changed.set()
        except Exception as e:
            logger.debug("reader stream %s error: %s", self._id, e)
            self._set_first_error(e)
            return

    async def _update_token_loop(self):
        while True:
            await asyncio.sleep(self._update_token_interval)
            token = self._get_token_function()
            if asyncio.iscoroutine(token):
                token = await token
            await self._update_token(token=token)

    async def _update_token(self, token: str):
        await self._update_token_event.wait()
        try:
            msg = StreamReadMessage.FromClient(UpdateTokenRequest(token))
            self._stream.write(msg)
        finally:
            self._update_token_event.clear()

    async def _on_start_partition_session(self, message: StreamReadMessage.StartPartitionSessionRequest):
        try:
            if message.partition_session.partition_session_id in self._partition_sessions:
                raise TopicReaderError(
                    "Double start partition session: %s" % message.partition_session.partition_session_id
                )

            self._partition_sessions[message.partition_session.partition_session_id] = datatypes.PartitionSession(
                id=message.partition_session.partition_session_id,
                state=datatypes.PartitionSession.State.Active,
                topic_path=message.partition_session.path,
                partition_id=message.partition_session.partition_id,
                committed_offset=message.committed_offset,
                reader_reconnector_id=self._reader_reconnector_id,
                reader_stream_id=self._id,
            )

            read_offset = None

            if self._settings.event_handler is not None:
                resp = await self._settings.event_handler._dispatch(
                    events.OnPartitionGetStartOffsetRequest(
                        message.partition_session.path,
                        message.partition_session.partition_id,
                    )
                )
                read_offset = None if resp is None else resp.start_offset

            self._stream.write(
                StreamReadMessage.FromClient(
                    client_message=StreamReadMessage.StartPartitionSessionResponse(
                        partition_session_id=message.partition_session.partition_session_id,
                        read_offset=read_offset,
                        commit_offset=None,
                    )
                ),
            )
        except YdbError as err:
            self._set_first_error(err)

    def _on_partition_session_stop(self, message: StreamReadMessage.StopPartitionSessionRequest):
        if message.partition_session_id not in self._partition_sessions:
            # may if receive stop partition with graceful=false after response on stop partition
            # with graceful=true and remove partition from internal dictionary
            return

        partition = self._partition_sessions.pop(message.partition_session_id)
        partition.close()

        if message.graceful:
            self._stream.write(
                StreamReadMessage.FromClient(
                    client_message=StreamReadMessage.StopPartitionSessionResponse(
                        partition_session_id=message.partition_session_id,
                    )
                )
            )

    def _on_end_partition_session(self, message: StreamReadMessage.EndPartitionSession):
        logger.debug(
            f"End partition session with id: {message.partition_session_id}, "
            f"child partitions: {message.child_partition_ids}"
        )

        if message.partition_session_id in self._partition_sessions:
            # Mark partition session as ended not to shuffle messages.
            self._partition_sessions[message.partition_session_id].end()

    def _on_read_response(self, message: StreamReadMessage.ReadResponse):
        self._buffer_consume_bytes(message.bytes_size)

        batches = self._read_response_to_batches(message)
        for batch in batches:
            self._batches_to_decode.put_nowait(batch)

    def _on_commit_response(self, message: StreamReadMessage.CommitOffsetResponse):
        for partition_offset in message.partitions_committed_offsets:
            if partition_offset.partition_session_id not in self._partition_sessions:
                continue

            session = self._partition_sessions[partition_offset.partition_session_id]
            session.ack_notify(partition_offset.committed_offset)

    def _buffer_consume_bytes(self, bytes_size):
        self._buffer_size_bytes -= bytes_size

    def _buffer_release_bytes(self, bytes_size):
        self._buffer_size_bytes += bytes_size
        self._stream.write(
            StreamReadMessage.FromClient(
                client_message=StreamReadMessage.ReadRequest(
                    bytes_size=bytes_size,
                )
            )
        )

    def _read_response_to_batches(self, message: StreamReadMessage.ReadResponse) -> typing.List[datatypes.PublicBatch]:
        batches = []

        batch_count = sum(len(p.batches) for p in message.partition_data)
        if batch_count == 0:
            return batches

        bytes_per_batch = message.bytes_size // batch_count
        additional_bytes_to_last_batch = message.bytes_size - bytes_per_batch * batch_count

        for partition_data in message.partition_data:
            partition_session = self._partition_sessions[partition_data.partition_session_id]
            for server_batch in partition_data.batches:
                messages = []
                for message_data in server_batch.message_data:
                    mess = datatypes.PublicMessage(
                        seqno=message_data.seq_no,
                        created_at=message_data.created_at,
                        message_group_id=message_data.message_group_id,
                        session_metadata=server_batch.write_session_meta,
                        offset=message_data.offset,
                        written_at=server_batch.written_at,
                        producer_id=server_batch.producer_id,
                        data=message_data.data,
                        metadata_items=message_data.metadata_items,
                        _partition_session=partition_session,
                        _commit_start_offset=partition_session._next_message_start_commit_offset,
                        _commit_end_offset=message_data.offset + 1,
                    )
                    messages.append(mess)
                    partition_session._next_message_start_commit_offset = mess._commit_end_offset

                if messages:
                    batch = datatypes.PublicBatch(
                        messages=messages,
                        _partition_session=partition_session,
                        _bytes_size=bytes_per_batch,
                        _codec=Codec(server_batch.codec),
                    )
                    batches.append(batch)

        batches[-1]._bytes_size += additional_bytes_to_last_batch
        return batches

    async def _decode_batches_loop(self):
        while True:
            batch = await self._batches_to_decode.get()
            logger.debug("reader stream %s decode batch %s messages", self._id, len(batch.messages))
            await self._decode_batch_inplace(batch)
            self._add_batch_to_queue(batch)
            self._state_changed.set()

    def _add_batch_to_queue(self, batch: datatypes.PublicBatch):
        part_sess_id = batch._partition_session.id
        if part_sess_id in self._message_batches:
            self._message_batches[part_sess_id]._extend(batch)
            logger.debug(
                "reader stream %s extend batch partition=%s size=%s",
                self._id,
                part_sess_id,
                len(batch.messages),
            )
            return

        self._message_batches[part_sess_id] = batch
        logger.debug(
            "reader stream %s new batch partition=%s size=%s",
            self._id,
            part_sess_id,
            len(batch.messages),
        )

    async def _decode_batch_inplace(self, batch):
        if batch._codec == Codec.CODEC_RAW:
            return

        try:
            decode_func = self._decoders[batch._codec]
        except KeyError:
            raise PublicTopicReaderUnexpectedCodecError("Receive message with unexpected codec: %s" % batch._codec)

        decode_data_futures = []
        for message in batch.messages:
            future = self._loop.run_in_executor(self._decode_executor, decode_func, message.data)
            decode_data_futures.append(future)

        decoded_data = await asyncio.gather(*decode_data_futures)
        for index, message in enumerate(batch.messages):
            message.data = decoded_data[index]

        batch._codec = Codec.CODEC_RAW

    def _set_first_error(self, err: YdbError):
        try:
            self._first_error.set_result(err)
            self._state_changed.set()
        except asyncio.InvalidStateError:
            # skip later set errors
            pass

    def _get_first_error(self) -> Optional[YdbError]:
        if self._first_error.done():
            return self._first_error.result()

    async def flush(self):
        futures = []
        for session in self._partition_sessions.values():
            futures.extend(w.future for w in session._ack_waiters)

        if futures:
            await asyncio.wait(futures)

    async def close(self, flush: bool):
        if self._closed:
            return

        self._closed = True
        logger.debug("reader stream %s close", self._id)

        if flush:
            await self.flush()

        self._set_first_error(TopicReaderStreamClosedError())
        self._state_changed.set()
        self._stream.close()

        for session in self._partition_sessions.values():
            session.close()
        self._partition_sessions.clear()

        for task in self._background_tasks:
            task.cancel()

        if self._background_tasks:
            await asyncio.wait(self._background_tasks)

        logger.debug("reader stream %s was closed", self._id)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/ydb_formats.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18protos/ydb_formats.proto\x12\x0bYdb.Formats\"$\n\x12\x41rrowBatchSettings\x12\x0e\n\x06schema\x18\x01 \x01(\x0c\"\xda\x01\n\x0b\x43svSettings\x12\x11\n\tskip_rows\x18\x01 \x01(\r\x12\x11\n\tdelimiter\x18\x02 \x01(\x0c\x12\x12\n\nnull_value\x18\x03 \x01(\x0c\x12\x0e\n\x06header\x18\x04 \x01(\x08\x12\x31\n\x07quoting\x18\x05 \x01(\x0b\x32 .Ydb.Formats.CsvSettings.Quoting\x1aN\n\x07Quoting\x12\x10\n\x08\x64isabled\x18\x01 \x01(\x08\x12\x12\n\nquote_char\x18\x02 \x01(\x0c\x12\x1d\n\x15\x64ouble_quote_disabled\x18\x03 \x01(\x08\"\xac\x02\n\x13\x41rrowFormatSettings\x12L\n\x11\x63ompression_codec\x18\x01 \x01(\x0b\x32\x31.Ydb.Formats.ArrowFormatSettings.CompressionCodec\x1a\xc6\x01\n\x10\x43ompressionCodec\x12\x44\n\x04type\x18\x01 \x01(\x0e\x32\x36.Ydb.Formats.ArrowFormatSettings.CompressionCodec.Type\x12\x12\n\x05level\x18\x02 \x01(\x05H\x00\x88\x01\x01\"N\n\x04Type\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\r\n\tTYPE_NONE\x10\x01\x12\r\n\tTYPE_ZSTD\x10\x02\x12\x12\n\x0eTYPE_LZ4_FRAME\x10\x03\x42\x08\n\x06_level\"!\n\x0f\x41rrowFormatMeta\x12\x0e\n\x06schema\x18\x01 \x01(\x0c\x42W\n\x16tech.ydb.proto.formatsZ:github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Formats\xf8\x01\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'protos.ydb_formats_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026tech.ydb.proto.formatsZ:github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Formats\370\001\001'
  _ARROWBATCHSETTINGS._serialized_start=41
  _ARROWBATCHSETTINGS._serialized_end=77
  _CSVSETTINGS._serialized_start=80
  _CSVSETTINGS._serialized_end=298
  _CSVSETTINGS_QUOTING._serialized_start=220
  _CSVSETTINGS_QUOTING._serialized_end=298
  _ARROWFORMATSETTINGS._serialized_start=301
  _ARROWFORMATSETTINGS._serialized_end=601
  _ARROWFORMATSETTINGS_COMPRESSIONCODEC._serialized_start=403
  _ARROWFORMATSETTINGS_COMPRESSIONCODEC._serialized_end=601
  _ARROWFORMATSETTINGS_COMPRESSIONCODEC_TYPE._serialized_start=513
  _ARROWFORMATSETTINGS_COMPRESSIONCODEC_TYPE._serialized_end=591
  _ARROWFORMATMETA._serialized_start=603
  _ARROWFORMATMETA._serialized_end=636
# @@protoc_insertion_point(module_scope)



from .scheme import (
    SchemeEntryType,
    _wrap_scheme_entry,
    _wrap_list_directory_response,
)
from ._apis import ydb_scheme


def test_wrap_scheme_entry():
    assert _wrap_scheme_entry(ydb_scheme.Entry(type=1)).type is SchemeEntryType.DIRECTORY
    assert _wrap_scheme_entry(ydb_scheme.Entry(type=17)).type is SchemeEntryType.TOPIC

    assert _wrap_scheme_entry(ydb_scheme.Entry()).type is SchemeEntryType.TYPE_UNSPECIFIED
    assert _wrap_scheme_entry(ydb_scheme.Entry(type=10)).type is SchemeEntryType.TYPE_UNSPECIFIED
    assert _wrap_scheme_entry(ydb_scheme.Entry(type=1001)).type is SchemeEntryType.TYPE_UNSPECIFIED


def test_wrap_list_directory_response():
    d = _wrap_list_directory_response(None, ydb_scheme.ListDirectoryResponse())
    assert d.type is SchemeEntryType.TYPE_UNSPECIFIED



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ydb._grpc.v3.draft.protos import ydb_keyvalue_pb2 as draft_dot_protos_dot_ydb__keyvalue__pb2


class KeyValueServiceStub(object):
    """KeyValue tablets provide a simple key-value storage in a low-overhead and easy-to-shoot-your-leg manner.
    To use KeyValue tablets in an efficient way one must be familiar with the design of both the KeyValue tablet
    and the Distributed Storage underneath it.

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateVolume = channel.unary_unary(
                '/Ydb.KeyValue.V1.KeyValueService/CreateVolume',
                request_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.CreateVolumeRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.CreateVolumeResponse.FromString,
                )
        self.DropVolume = channel.unary_unary(
                '/Ydb.KeyValue.V1.KeyValueService/DropVolume',
                request_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.DropVolumeRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.DropVolumeResponse.FromString,
                )
        self.AlterVolume = channel.unary_unary(
                '/Ydb.KeyValue.V1.KeyValueService/AlterVolume',
                request_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.AlterVolumeRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.AlterVolumeResponse.FromString,
                )
        self.DescribeVolume = channel.unary_unary(
                '/Ydb.KeyValue.V1.KeyValueService/DescribeVolume',
                request_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.DescribeVolumeRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.DescribeVolumeResponse.FromString,
                )
        self.ListLocalPartitions = channel.unary_unary(
                '/Ydb.KeyValue.V1.KeyValueService/ListLocalPartitions',
                request_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ListLocalPartitionsRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ListLocalPartitionsResponse.FromString,
                )
        self.AcquireLock = channel.unary_unary(
                '/Ydb.KeyValue.V1.KeyValueService/AcquireLock',
                request_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.AcquireLockRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.AcquireLockResponse.FromString,
                )
        self.ExecuteTransaction = channel.unary_unary(
                '/Ydb.KeyValue.V1.KeyValueService/ExecuteTransaction',
                request_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ExecuteTransactionRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ExecuteTransactionResponse.FromString,
                )
        self.Read = channel.unary_unary(
                '/Ydb.KeyValue.V1.KeyValueService/Read',
                request_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ReadRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ReadResponse.FromString,
                )
        self.ReadRange = channel.unary_unary(
                '/Ydb.KeyValue.V1.KeyValueService/ReadRange',
                request_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ReadRangeRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ReadRangeResponse.FromString,
                )
        self.ListRange = channel.unary_unary(
                '/Ydb.KeyValue.V1.KeyValueService/ListRange',
                request_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ListRangeRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ListRangeResponse.FromString,
                )
        self.GetStorageChannelStatus = channel.unary_unary(
                '/Ydb.KeyValue.V1.KeyValueService/GetStorageChannelStatus',
                request_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.GetStorageChannelStatusRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.GetStorageChannelStatusResponse.FromString,
                )


class KeyValueServiceServicer(object):
    """KeyValue tablets provide a simple key-value storage in a low-overhead and easy-to-shoot-your-leg manner.
    To use KeyValue tablets in an efficient way one must be familiar with the design of both the KeyValue tablet
    and the Distributed Storage underneath it.

    """

    def CreateVolume(self, request, context):
        """Create a volume by path and partition count
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DropVolume(self, request, context):
        """Drop the volume by path
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AlterVolume(self, request, context):
        """Alter the volume by path
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DescribeVolume(self, request, context):
        """Describe the volume by path
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListLocalPartitions(self, request, context):
        """List partitions of a volume at the local node.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AcquireLock(self, request, context):
        """Acquire an exclusive lock for the partition.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExecuteTransaction(self, request, context):
        """Perform list of commands to modify the state of the partition as an atomic transaction.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Read(self, request, context):
        """Read the value stored in the item with the key specified.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReadRange(self, request, context):
        """Read items with keys in the specified range.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListRange(self, request, context):
        """List keys and metadata of items with keys in the specified range.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStorageChannelStatus(self, request, context):
        """Get storage channel status of the partition.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_KeyValueServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreateVolume': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateVolume,
                    request_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.CreateVolumeRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.CreateVolumeResponse.SerializeToString,
            ),
            'DropVolume': grpc.unary_unary_rpc_method_handler(
                    servicer.DropVolume,
                    request_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.DropVolumeRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.DropVolumeResponse.SerializeToString,
            ),
            'AlterVolume': grpc.unary_unary_rpc_method_handler(
                    servicer.AlterVolume,
                    request_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.AlterVolumeRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.AlterVolumeResponse.SerializeToString,
            ),
            'DescribeVolume': grpc.unary_unary_rpc_method_handler(
                    servicer.DescribeVolume,
                    request_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.DescribeVolumeRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.DescribeVolumeResponse.SerializeToString,
            ),
            'ListLocalPartitions': grpc.unary_unary_rpc_method_handler(
                    servicer.ListLocalPartitions,
                    request_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ListLocalPartitionsRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ListLocalPartitionsResponse.SerializeToString,
            ),
            'AcquireLock': grpc.unary_unary_rpc_method_handler(
                    servicer.AcquireLock,
                    request_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.AcquireLockRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.AcquireLockResponse.SerializeToString,
            ),
            'ExecuteTransaction': grpc.unary_unary_rpc_method_handler(
                    servicer.ExecuteTransaction,
                    request_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ExecuteTransactionRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ExecuteTransactionResponse.SerializeToString,
            ),
            'Read': grpc.unary_unary_rpc_method_handler(
                    servicer.Read,
                    request_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ReadRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ReadResponse.SerializeToString,
            ),
            'ReadRange': grpc.unary_unary_rpc_method_handler(
                    servicer.ReadRange,
                    request_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ReadRangeRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ReadRangeResponse.SerializeToString,
            ),
            'ListRange': grpc.unary_unary_rpc_method_handler(
                    servicer.ListRange,
                    request_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ListRangeRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.ListRangeResponse.SerializeToString,
            ),
            'GetStorageChannelStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStorageChannelStatus,
                    request_deserializer=draft_dot_protos_dot_ydb__keyvalue__pb2.GetStorageChannelStatusRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__keyvalue__pb2.GetStorageChannelStatusResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Ydb.KeyValue.V1.KeyValueService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class KeyValueService(object):
    """KeyValue tablets provide a simple key-value storage in a low-overhead and easy-to-shoot-your-leg manner.
    To use KeyValue tablets in an efficient way one must be familiar with the design of both the KeyValue tablet
    and the Distributed Storage underneath it.

    """

    @staticmethod
    def CreateVolume(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.KeyValue.V1.KeyValueService/CreateVolume',
            draft_dot_protos_dot_ydb__keyvalue__pb2.CreateVolumeRequest.SerializeToString,
            draft_dot_protos_dot_ydb__keyvalue__pb2.CreateVolumeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DropVolume(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.KeyValue.V1.KeyValueService/DropVolume',
            draft_dot_protos_dot_ydb__keyvalue__pb2.DropVolumeRequest.SerializeToString,
            draft_dot_protos_dot_ydb__keyvalue__pb2.DropVolumeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AlterVolume(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.KeyValue.V1.KeyValueService/AlterVolume',
            draft_dot_protos_dot_ydb__keyvalue__pb2.AlterVolumeRequest.SerializeToString,
            draft_dot_protos_dot_ydb__keyvalue__pb2.AlterVolumeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DescribeVolume(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.KeyValue.V1.KeyValueService/DescribeVolume',
            draft_dot_protos_dot_ydb__keyvalue__pb2.DescribeVolumeRequest.SerializeToString,
            draft_dot_protos_dot_ydb__keyvalue__pb2.DescribeVolumeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListLocalPartitions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.KeyValue.V1.KeyValueService/ListLocalPartitions',
            draft_dot_protos_dot_ydb__keyvalue__pb2.ListLocalPartitionsRequest.SerializeToString,
            draft_dot_protos_dot_ydb__keyvalue__pb2.ListLocalPartitionsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AcquireLock(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.KeyValue.V1.KeyValueService/AcquireLock',
            draft_dot_protos_dot_ydb__keyvalue__pb2.AcquireLockRequest.SerializeToString,
            draft_dot_protos_dot_ydb__keyvalue__pb2.AcquireLockResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ExecuteTransaction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.KeyValue.V1.KeyValueService/ExecuteTransaction',
            draft_dot_protos_dot_ydb__keyvalue__pb2.ExecuteTransactionRequest.SerializeToString,
            draft_dot_protos_dot_ydb__keyvalue__pb2.ExecuteTransactionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Read(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.KeyValue.V1.KeyValueService/Read',
            draft_dot_protos_dot_ydb__keyvalue__pb2.ReadRequest.SerializeToString,
            draft_dot_protos_dot_ydb__keyvalue__pb2.ReadResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ReadRange(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.KeyValue.V1.KeyValueService/ReadRange',
            draft_dot_protos_dot_ydb__keyvalue__pb2.ReadRangeRequest.SerializeToString,
            draft_dot_protos_dot_ydb__keyvalue__pb2.ReadRangeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListRange(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.KeyValue.V1.KeyValueService/ListRange',
            draft_dot_protos_dot_ydb__keyvalue__pb2.ListRangeRequest.SerializeToString,
            draft_dot_protos_dot_ydb__keyvalue__pb2.ListRangeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetStorageChannelStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.KeyValue.V1.KeyValueService/GetStorageChannelStatus',
            draft_dot_protos_dot_ydb__keyvalue__pb2.GetStorageChannelStatusRequest.SerializeToString,
            draft_dot_protos_dot_ydb__keyvalue__pb2.GetStorageChannelStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import logging
from concurrent import futures
from typing import (
    Callable,
    Optional,
    List,
    Dict,
    Any,
    Union,
)
import time
import threading
import queue

from .base import BaseQueryTxMode, QueryExplainResultFormat
from .base import QueryClientSettings
from .session import (
    QuerySession,
)
from ..retries import (
    RetrySettings,
    retry_operation_sync,
)
from .. import issues
from .. import convert
from ..settings import BaseRequestSettings
from .._grpc.grpcwrapper import common_utils
from .._grpc.grpcwrapper import ydb_query_public_types as _ydb_query_public


logger = logging.getLogger(__name__)


class QuerySessionPool:
    """QuerySessionPool is an object to simplify operations with sessions of Query Service."""

    def __init__(
        self,
        driver: common_utils.SupportedDriverType,
        size: int = 100,
        *,
        query_client_settings: Optional[QueryClientSettings] = None,
        workers_threads_count: int = 4,
    ):
        """
        :param driver: A driver instance.
        :param size: Max size of Session Pool.
        :param query_client_settings: ydb.QueryClientSettings object to configure QueryService behavior
        :param workers_threads_count: A number of threads in executor used for *_async methods
        """

        self._driver = driver
        self._tp = futures.ThreadPoolExecutor(workers_threads_count)
        self._queue = queue.Queue()
        self._current_size = 0
        self._size = size
        self._should_stop = threading.Event()
        self._lock = threading.RLock()
        self._query_client_settings = query_client_settings

    def _create_new_session(self, timeout: Optional[float]):
        session = QuerySession(self._driver, settings=self._query_client_settings)
        session.create(settings=BaseRequestSettings().with_timeout(timeout))
        logger.debug(f"New session was created for pool. Session id: {session.session_id}")
        return session

    def acquire(self, timeout: Optional[float] = None) -> QuerySession:
        """Acquire a session from Session Pool.

        :param timeout: A timeout to wait in seconds.

        :return A QuerySession object.
        """

        start = time.monotonic()

        lock_acquire_timeout = timeout if timeout is not None else -1
        acquired = self._lock.acquire(timeout=lock_acquire_timeout)
        try:
            if self._should_stop.is_set():
                logger.error("An attempt to take session from closed session pool.")
                raise issues.SessionPoolClosed()

            session = None
            try:
                session = self._queue.get_nowait()
            except queue.Empty:
                pass

            finish = time.monotonic()
            timeout = timeout - (finish - start) if timeout is not None else None

            start = time.monotonic()
            if session is None and self._current_size == self._size:
                try:
                    session = self._queue.get(block=True, timeout=timeout)
                except queue.Empty:
                    raise issues.SessionPoolEmpty("Timeout on acquire session")

            if session is not None:
                if session.is_active:
                    logger.debug(f"Acquired active session from queue: {session.session_id}")
                    return session
                else:
                    self._current_size -= 1
                    logger.debug(f"Acquired dead session from queue: {session.session_id}")

            logger.debug(f"Session pool is not large enough: {self._current_size} < {self._size}, will create new one.")
            finish = time.monotonic()
            time_left = timeout - (finish - start) if timeout is not None else None
            session = self._create_new_session(time_left)

            self._current_size += 1
            return session
        finally:
            if acquired:
                self._lock.release()

    def release(self, session: QuerySession) -> None:
        """Release a session back to Session Pool."""

        self._queue.put_nowait(session)
        logger.debug("Session returned to queue: %s", session.session_id)

    def checkout(self, timeout: Optional[float] = None) -> "SimpleQuerySessionCheckout":
        """Return a Session context manager, that acquires session on enter and releases session on exit.

        :param timeout: A timeout to wait in seconds.
        """

        return SimpleQuerySessionCheckout(self, timeout)

    def retry_operation_sync(self, callee: Callable, retry_settings: Optional[RetrySettings] = None, *args, **kwargs):
        """Special interface to execute a bunch of commands with session in a safe, retriable way.

        :param callee: A function, that works with session.
        :param retry_settings: RetrySettings object.

        :return: Result sets or exception in case of execution errors.
        """

        if self._should_stop.is_set():
            raise issues.SessionPoolClosed()

        retry_settings = RetrySettings() if retry_settings is None else retry_settings

        def wrapped_callee():
            with self.checkout(timeout=retry_settings.max_session_acquire_timeout) as session:
                return callee(session, *args, **kwargs)

        return retry_operation_sync(wrapped_callee, retry_settings)

    def retry_tx_async(
        self,
        callee: Callable,
        tx_mode: Optional[BaseQueryTxMode] = None,
        retry_settings: Optional[RetrySettings] = None,
        *args,
        **kwargs,
    ) -> futures.Future:
        """Asynchronously execute a transaction in a retriable way."""

        if self._should_stop.is_set():
            raise issues.SessionPoolClosed()

        return self._tp.submit(
            self.retry_tx_sync,
            callee,
            tx_mode,
            retry_settings,
            *args,
            **kwargs,
        )

    def retry_operation_async(
        self, callee: Callable, retry_settings: Optional[RetrySettings] = None, *args, **kwargs
    ) -> futures.Future:
        """Asynchronously execute a retryable operation."""

        if self._should_stop.is_set():
            raise issues.SessionPoolClosed()

        return self._tp.submit(self.retry_operation_sync, callee, retry_settings, *args, **kwargs)

    def retry_tx_sync(
        self,
        callee: Callable,
        tx_mode: Optional[BaseQueryTxMode] = None,
        retry_settings: Optional[RetrySettings] = None,
        *args,
        **kwargs,
    ):
        """Special interface to execute a bunch of commands with transaction in a safe, retriable way.

        :param callee: A function, that works with session.
        :param tx_mode: Transaction mode, which is a one from the following choices:
          1) QuerySerializableReadWrite() which is default mode;
          2) QueryOnlineReadOnly(allow_inconsistent_reads=False);
          3) QuerySnapshotReadOnly();
          4) QuerySnapshotReadWrite();
          5) QueryStaleReadOnly().
        :param retry_settings: RetrySettings object.

        :return: Result sets or exception in case of execution errors.
        """

        if self._should_stop.is_set():
            raise issues.SessionPoolClosed()

        tx_mode = tx_mode if tx_mode else _ydb_query_public.QuerySerializableReadWrite()
        retry_settings = RetrySettings() if retry_settings is None else retry_settings

        def wrapped_callee():
            with self.checkout(timeout=retry_settings.max_session_acquire_timeout) as session:
                with session.transaction(tx_mode=tx_mode) as tx:
                    if tx_mode.name in ["serializable_read_write", "snapshot_read_only"]:
                        tx.begin()
                    result = callee(tx, *args, **kwargs)
                    tx.commit()
                return result

        return retry_operation_sync(wrapped_callee, retry_settings)

    def execute_with_retries(
        self,
        query: str,
        parameters: Optional[dict] = None,
        retry_settings: Optional[RetrySettings] = None,
        *args,
        **kwargs,
    ) -> List[convert.ResultSet]:
        """Special interface to execute a one-shot queries in a safe, retriable way.
        Note: this method loads all data from stream before return, do not use this
        method with huge read queries.

        :param query: A query, yql or sql text.
        :param parameters: dict with parameters and YDB types;
        :param retry_settings: RetrySettings object.

        :return: Result sets or exception in case of execution errors.
        """

        if self._should_stop.is_set():
            raise issues.SessionPoolClosed()

        retry_settings = RetrySettings() if retry_settings is None else retry_settings

        def wrapped_callee():
            with self.checkout(timeout=retry_settings.max_session_acquire_timeout) as session:
                it = session.execute(query, parameters, *args, **kwargs)
                return [result_set for result_set in it]

        return retry_operation_sync(wrapped_callee, retry_settings)

    def execute_with_retries_async(
        self,
        query: str,
        parameters: Optional[dict] = None,
        retry_settings: Optional[RetrySettings] = None,
        *args,
        **kwargs,
    ) -> futures.Future:
        """Asynchronously execute a query with retries."""

        if self._should_stop.is_set():
            raise issues.SessionPoolClosed()

        return self._tp.submit(
            self.execute_with_retries,
            query,
            parameters,
            retry_settings,
            *args,
            **kwargs,
        )

    def explain_with_retries(
        self,
        query: str,
        parameters: Optional[dict] = None,
        *,
        result_format: QueryExplainResultFormat = QueryExplainResultFormat.STR,
        retry_settings: Optional[RetrySettings] = None,
    ) -> Union[str, Dict[str, Any]]:
        """
        Explain a query in retriable way. No real query execution will happen.

        :param query: A query, yql or sql text.
        :param parameters: dict with parameters and YDB types;
        :param result_format: Return format: string or dict.
        :param retry_settings: RetrySettings object.
        :return: Parsed query plan.
        """

        def callee(session: QuerySession):
            return session.explain(query, parameters, result_format=result_format)

        return self.retry_operation_sync(callee, retry_settings)

    def stop(self, timeout=None):
        acquire_timeout = timeout if timeout is not None else -1
        acquired = self._lock.acquire(timeout=acquire_timeout)
        try:
            self._should_stop.set()
            self._tp.shutdown(wait=True)
            while True:
                try:
                    session = self._queue.get_nowait()
                    session.delete()
                except queue.Empty:
                    break

            logger.debug("All session were deleted.")
        finally:
            if acquired:
                self._lock.release()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


class SimpleQuerySessionCheckout:
    def __init__(self, pool: QuerySessionPool, timeout: Optional[float]):
        self._pool = pool
        self._timeout = timeout
        self._session = None

    def __enter__(self) -> QuerySession:
        self._session = self._pool.acquire(self._timeout)
        return self._session

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._pool.release(self._session)



from .. import _apis, issues
from .._grpc.grpcwrapper.ydb_coordination_public_types import NodeConfig, DescribeResult
import logging

logger = logging.getLogger(__name__)


def wrapper_create_node(rpc_state, response_pb):
    issues._process_response(response_pb.operation)


def wrapper_describe_node(rpc_state, response_pb) -> NodeConfig:
    issues._process_response(response_pb.operation)
    return DescribeResult.from_proto(response_pb)


def wrapper_delete_node(rpc_state, response_pb):
    issues._process_response(response_pb.operation)


def wrapper_alter_node(rpc_state, response_pb):
    issues._process_response(response_pb.operation)


class BaseCoordinationClient:
    def __init__(self, driver):
        self._driver = driver
        self._user_warned = False

    def _call_create(self, request, settings=None):
        return self._driver(
            request,
            _apis.CoordinationService.Stub,
            _apis.CoordinationService.CreateNode,
            wrap_result=wrapper_create_node,
            settings=settings,
        )

    def _call_describe(self, request, settings=None):
        return self._driver(
            request,
            _apis.CoordinationService.Stub,
            _apis.CoordinationService.DescribeNode,
            wrap_result=wrapper_describe_node,
            settings=settings,
        )

    def _call_alter(self, request, settings=None):
        return self._driver(
            request,
            _apis.CoordinationService.Stub,
            _apis.CoordinationService.AlterNode,
            wrap_result=wrapper_alter_node,
            settings=settings,
        )

    def _call_delete(self, request, settings=None):
        return self._driver(
            request,
            _apis.CoordinationService.Stub,
            _apis.CoordinationService.DropNode,
            wrap_result=wrapper_delete_node,
            settings=settings,
        )

    def _log_experimental_api(self):
        if not self._user_warned:
            logger.warning(
                "Coordination Service API is experimental, may contain bugs and may change in future releases."
            )
            self._user_warned = True



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ydb._grpc.v5.protos import ydb_auth_pb2 as protos_dot_ydb__auth__pb2


class AuthServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Login = channel.unary_unary(
                '/Ydb.Auth.V1.AuthService/Login',
                request_serializer=protos_dot_ydb__auth__pb2.LoginRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__auth__pb2.LoginResponse.FromString,
                )


class AuthServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Login(self, request, context):
        """Perform login using built-in auth system
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AuthServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Login': grpc.unary_unary_rpc_method_handler(
                    servicer.Login,
                    request_deserializer=protos_dot_ydb__auth__pb2.LoginRequest.FromString,
                    response_serializer=protos_dot_ydb__auth__pb2.LoginResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Ydb.Auth.V1.AuthService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class AuthService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Login(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Auth.V1.AuthService/Login',
            protos_dot_ydb__auth__pb2.LoginRequest.SerializeToString,
            protos_dot_ydb__auth__pb2.LoginResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_federation_discovery_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.protos import ydb_federation_discovery_pb2 as protos_dot_ydb__federation__discovery__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ydb_federation_discovery_v1.proto',
  package='Ydb.FederationDiscovery.V1',
  syntax='proto3',
  serialized_options=b'\n&tech.ydb.proto.federation.discovery.v1ZBgithub.com/ydb-platform/ydb-go-genproto/Ydb_FederationDiscovery_V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n!ydb_federation_discovery_v1.proto\x12\x1aYdb.FederationDiscovery.V1\x1a%protos/ydb_federation_discovery.proto2\xab\x01\n\x1a\x46\x65\x64\x65rationDiscoveryService\x12\x8c\x01\n\x17ListFederationDatabases\x12\x37.Ydb.FederationDiscovery.ListFederationDatabasesRequest\x1a\x38.Ydb.FederationDiscovery.ListFederationDatabasesResponseBl\n&tech.ydb.proto.federation.discovery.v1ZBgithub.com/ydb-platform/ydb-go-genproto/Ydb_FederationDiscovery_V1b\x06proto3'
  ,
  dependencies=[protos_dot_ydb__federation__discovery__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_FEDERATIONDISCOVERYSERVICE = _descriptor.ServiceDescriptor(
  name='FederationDiscoveryService',
  full_name='Ydb.FederationDiscovery.V1.FederationDiscoveryService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=105,
  serialized_end=276,
  methods=[
  _descriptor.MethodDescriptor(
    name='ListFederationDatabases',
    full_name='Ydb.FederationDiscovery.V1.FederationDiscoveryService.ListFederationDatabases',
    index=0,
    containing_service=None,
    input_type=protos_dot_ydb__federation__discovery__pb2._LISTFEDERATIONDATABASESREQUEST,
    output_type=protos_dot_ydb__federation__discovery__pb2._LISTFEDERATIONDATABASESRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_FEDERATIONDISCOVERYSERVICE)

DESCRIPTOR.services_by_name['FederationDiscoveryService'] = _FEDERATIONDISCOVERYSERVICE

# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ydb._grpc.v5.protos import ydb_cms_pb2 as protos_dot_ydb__cms__pb2


class CmsServiceStub(object):
    """CMS stands for Cluster Management System. CmsService provides some
    functionality for managing cluster, i.e. managing YDB Database
    instances for example.

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateDatabase = channel.unary_unary(
                '/Ydb.Cms.V1.CmsService/CreateDatabase',
                request_serializer=protos_dot_ydb__cms__pb2.CreateDatabaseRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__cms__pb2.CreateDatabaseResponse.FromString,
                )
        self.GetDatabaseStatus = channel.unary_unary(
                '/Ydb.Cms.V1.CmsService/GetDatabaseStatus',
                request_serializer=protos_dot_ydb__cms__pb2.GetDatabaseStatusRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__cms__pb2.GetDatabaseStatusResponse.FromString,
                )
        self.AlterDatabase = channel.unary_unary(
                '/Ydb.Cms.V1.CmsService/AlterDatabase',
                request_serializer=protos_dot_ydb__cms__pb2.AlterDatabaseRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__cms__pb2.AlterDatabaseResponse.FromString,
                )
        self.ListDatabases = channel.unary_unary(
                '/Ydb.Cms.V1.CmsService/ListDatabases',
                request_serializer=protos_dot_ydb__cms__pb2.ListDatabasesRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__cms__pb2.ListDatabasesResponse.FromString,
                )
        self.RemoveDatabase = channel.unary_unary(
                '/Ydb.Cms.V1.CmsService/RemoveDatabase',
                request_serializer=protos_dot_ydb__cms__pb2.RemoveDatabaseRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__cms__pb2.RemoveDatabaseResponse.FromString,
                )
        self.DescribeDatabaseOptions = channel.unary_unary(
                '/Ydb.Cms.V1.CmsService/DescribeDatabaseOptions',
                request_serializer=protos_dot_ydb__cms__pb2.DescribeDatabaseOptionsRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__cms__pb2.DescribeDatabaseOptionsResponse.FromString,
                )


class CmsServiceServicer(object):
    """CMS stands for Cluster Management System. CmsService provides some
    functionality for managing cluster, i.e. managing YDB Database
    instances for example.

    """

    def CreateDatabase(self, request, context):
        """Create a new database.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDatabaseStatus(self, request, context):
        """Get current database's status.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AlterDatabase(self, request, context):
        """Alter database resources.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListDatabases(self, request, context):
        """List all databases.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemoveDatabase(self, request, context):
        """Remove database.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DescribeDatabaseOptions(self, request, context):
        """Describe supported database options.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_CmsServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreateDatabase': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateDatabase,
                    request_deserializer=protos_dot_ydb__cms__pb2.CreateDatabaseRequest.FromString,
                    response_serializer=protos_dot_ydb__cms__pb2.CreateDatabaseResponse.SerializeToString,
            ),
            'GetDatabaseStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDatabaseStatus,
                    request_deserializer=protos_dot_ydb__cms__pb2.GetDatabaseStatusRequest.FromString,
                    response_serializer=protos_dot_ydb__cms__pb2.GetDatabaseStatusResponse.SerializeToString,
            ),
            'AlterDatabase': grpc.unary_unary_rpc_method_handler(
                    servicer.AlterDatabase,
                    request_deserializer=protos_dot_ydb__cms__pb2.AlterDatabaseRequest.FromString,
                    response_serializer=protos_dot_ydb__cms__pb2.AlterDatabaseResponse.SerializeToString,
            ),
            'ListDatabases': grpc.unary_unary_rpc_method_handler(
                    servicer.ListDatabases,
                    request_deserializer=protos_dot_ydb__cms__pb2.ListDatabasesRequest.FromString,
                    response_serializer=protos_dot_ydb__cms__pb2.ListDatabasesResponse.SerializeToString,
            ),
            'RemoveDatabase': grpc.unary_unary_rpc_method_handler(
                    servicer.RemoveDatabase,
                    request_deserializer=protos_dot_ydb__cms__pb2.RemoveDatabaseRequest.FromString,
                    response_serializer=protos_dot_ydb__cms__pb2.RemoveDatabaseResponse.SerializeToString,
            ),
            'DescribeDatabaseOptions': grpc.unary_unary_rpc_method_handler(
                    servicer.DescribeDatabaseOptions,
                    request_deserializer=protos_dot_ydb__cms__pb2.DescribeDatabaseOptionsRequest.FromString,
                    response_serializer=protos_dot_ydb__cms__pb2.DescribeDatabaseOptionsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Ydb.Cms.V1.CmsService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class CmsService(object):
    """CMS stands for Cluster Management System. CmsService provides some
    functionality for managing cluster, i.e. managing YDB Database
    instances for example.

    """

    @staticmethod
    def CreateDatabase(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Cms.V1.CmsService/CreateDatabase',
            protos_dot_ydb__cms__pb2.CreateDatabaseRequest.SerializeToString,
            protos_dot_ydb__cms__pb2.CreateDatabaseResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetDatabaseStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Cms.V1.CmsService/GetDatabaseStatus',
            protos_dot_ydb__cms__pb2.GetDatabaseStatusRequest.SerializeToString,
            protos_dot_ydb__cms__pb2.GetDatabaseStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AlterDatabase(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Cms.V1.CmsService/AlterDatabase',
            protos_dot_ydb__cms__pb2.AlterDatabaseRequest.SerializeToString,
            protos_dot_ydb__cms__pb2.AlterDatabaseResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListDatabases(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Cms.V1.CmsService/ListDatabases',
            protos_dot_ydb__cms__pb2.ListDatabasesRequest.SerializeToString,
            protos_dot_ydb__cms__pb2.ListDatabasesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RemoveDatabase(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Cms.V1.CmsService/RemoveDatabase',
            protos_dot_ydb__cms__pb2.RemoveDatabaseRequest.SerializeToString,
            protos_dot_ydb__cms__pb2.RemoveDatabaseResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DescribeDatabaseOptions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Cms.V1.CmsService/DescribeDatabaseOptions',
            protos_dot_ydb__cms__pb2.DescribeDatabaseOptionsRequest.SerializeToString,
            protos_dot_ydb__cms__pb2.DescribeDatabaseOptionsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ydb._grpc.v5.protos import ydb_rate_limiter_pb2 as protos_dot_ydb__rate__limiter__pb2


class RateLimiterServiceStub(object):
    """Service that implements distributed rate limiting.

    To use rate limiter functionality you need an existing coordination node.

    Control plane API
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateResource = channel.unary_unary(
                '/Ydb.RateLimiter.V1.RateLimiterService/CreateResource',
                request_serializer=protos_dot_ydb__rate__limiter__pb2.CreateResourceRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__rate__limiter__pb2.CreateResourceResponse.FromString,
                )
        self.AlterResource = channel.unary_unary(
                '/Ydb.RateLimiter.V1.RateLimiterService/AlterResource',
                request_serializer=protos_dot_ydb__rate__limiter__pb2.AlterResourceRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__rate__limiter__pb2.AlterResourceResponse.FromString,
                )
        self.DropResource = channel.unary_unary(
                '/Ydb.RateLimiter.V1.RateLimiterService/DropResource',
                request_serializer=protos_dot_ydb__rate__limiter__pb2.DropResourceRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__rate__limiter__pb2.DropResourceResponse.FromString,
                )
        self.ListResources = channel.unary_unary(
                '/Ydb.RateLimiter.V1.RateLimiterService/ListResources',
                request_serializer=protos_dot_ydb__rate__limiter__pb2.ListResourcesRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__rate__limiter__pb2.ListResourcesResponse.FromString,
                )
        self.DescribeResource = channel.unary_unary(
                '/Ydb.RateLimiter.V1.RateLimiterService/DescribeResource',
                request_serializer=protos_dot_ydb__rate__limiter__pb2.DescribeResourceRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__rate__limiter__pb2.DescribeResourceResponse.FromString,
                )
        self.AcquireResource = channel.unary_unary(
                '/Ydb.RateLimiter.V1.RateLimiterService/AcquireResource',
                request_serializer=protos_dot_ydb__rate__limiter__pb2.AcquireResourceRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__rate__limiter__pb2.AcquireResourceResponse.FromString,
                )


class RateLimiterServiceServicer(object):
    """Service that implements distributed rate limiting.

    To use rate limiter functionality you need an existing coordination node.

    Control plane API
    """

    def CreateResource(self, request, context):
        """Create a new resource in existing coordination node.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AlterResource(self, request, context):
        """Update a resource in coordination node.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DropResource(self, request, context):
        """Delete a resource from coordination node.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListResources(self, request, context):
        """List resources in given coordination node.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DescribeResource(self, request, context):
        """Describe properties of resource in coordination node.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AcquireResource(self, request, context):
        """Take units for usage of a resource in coordination node.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RateLimiterServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreateResource': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateResource,
                    request_deserializer=protos_dot_ydb__rate__limiter__pb2.CreateResourceRequest.FromString,
                    response_serializer=protos_dot_ydb__rate__limiter__pb2.CreateResourceResponse.SerializeToString,
            ),
            'AlterResource': grpc.unary_unary_rpc_method_handler(
                    servicer.AlterResource,
                    request_deserializer=protos_dot_ydb__rate__limiter__pb2.AlterResourceRequest.FromString,
                    response_serializer=protos_dot_ydb__rate__limiter__pb2.AlterResourceResponse.SerializeToString,
            ),
            'DropResource': grpc.unary_unary_rpc_method_handler(
                    servicer.DropResource,
                    request_deserializer=protos_dot_ydb__rate__limiter__pb2.DropResourceRequest.FromString,
                    response_serializer=protos_dot_ydb__rate__limiter__pb2.DropResourceResponse.SerializeToString,
            ),
            'ListResources': grpc.unary_unary_rpc_method_handler(
                    servicer.ListResources,
                    request_deserializer=protos_dot_ydb__rate__limiter__pb2.ListResourcesRequest.FromString,
                    response_serializer=protos_dot_ydb__rate__limiter__pb2.ListResourcesResponse.SerializeToString,
            ),
            'DescribeResource': grpc.unary_unary_rpc_method_handler(
                    servicer.DescribeResource,
                    request_deserializer=protos_dot_ydb__rate__limiter__pb2.DescribeResourceRequest.FromString,
                    response_serializer=protos_dot_ydb__rate__limiter__pb2.DescribeResourceResponse.SerializeToString,
            ),
            'AcquireResource': grpc.unary_unary_rpc_method_handler(
                    servicer.AcquireResource,
                    request_deserializer=protos_dot_ydb__rate__limiter__pb2.AcquireResourceRequest.FromString,
                    response_serializer=protos_dot_ydb__rate__limiter__pb2.AcquireResourceResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Ydb.RateLimiter.V1.RateLimiterService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class RateLimiterService(object):
    """Service that implements distributed rate limiting.

    To use rate limiter functionality you need an existing coordination node.

    Control plane API
    """

    @staticmethod
    def CreateResource(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.RateLimiter.V1.RateLimiterService/CreateResource',
            protos_dot_ydb__rate__limiter__pb2.CreateResourceRequest.SerializeToString,
            protos_dot_ydb__rate__limiter__pb2.CreateResourceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AlterResource(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.RateLimiter.V1.RateLimiterService/AlterResource',
            protos_dot_ydb__rate__limiter__pb2.AlterResourceRequest.SerializeToString,
            protos_dot_ydb__rate__limiter__pb2.AlterResourceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DropResource(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.RateLimiter.V1.RateLimiterService/DropResource',
            protos_dot_ydb__rate__limiter__pb2.DropResourceRequest.SerializeToString,
            protos_dot_ydb__rate__limiter__pb2.DropResourceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListResources(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.RateLimiter.V1.RateLimiterService/ListResources',
            protos_dot_ydb__rate__limiter__pb2.ListResourcesRequest.SerializeToString,
            protos_dot_ydb__rate__limiter__pb2.ListResourcesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DescribeResource(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.RateLimiter.V1.RateLimiterService/DescribeResource',
            protos_dot_ydb__rate__limiter__pb2.DescribeResourceRequest.SerializeToString,
            protos_dot_ydb__rate__limiter__pb2.DescribeResourceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AcquireResource(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.RateLimiter.V1.RateLimiterService/AcquireResource',
            protos_dot_ydb__rate__limiter__pb2.AcquireResourceRequest.SerializeToString,
            protos_dot_ydb__rate__limiter__pb2.AcquireResourceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import asyncio

from ydb.aio.connection import Connection
import pytest
from ydb import _apis


@pytest.mark.asyncio
async def test_async_call(aio_connection):
    request = _apis.ydb_scheme.DescribePathRequest()
    request.path = ""
    stub = _apis.SchemeService.Stub
    response = await aio_connection(request, stub, _apis.SchemeService.DescribePath)
    assert response.operation.ready


@pytest.mark.asyncio
async def test_close(aio_connection: Connection):
    request = _apis.ydb_scheme.DescribePathRequest()
    request.path = ""
    stub = _apis.SchemeService.Stub

    async def run_and_catch():
        from ydb.issues import ConnectionLost

        try:
            await aio_connection(request, stub, _apis.SchemeService.DescribePath)
        except ConnectionLost:
            pass

    tasks = [run_and_catch() for _ in range(10)]

    tasks.append(aio_connection.close(0))
    with pytest.raises(asyncio.CancelledError):
        await asyncio.gather(
            *tasks,
        )


@pytest.mark.asyncio
async def test_callbacks(aio_connection: Connection):
    cleanup_callback_called = False
    result_wrapped = False
    wrapper_args = ("1", "3", 10, "239")

    request = _apis.ydb_scheme.DescribePathRequest()
    request.path = ""
    stub = _apis.SchemeService.Stub

    def cleanup(conn):
        assert conn is aio_connection
        nonlocal cleanup_callback_called
        cleanup_callback_called = True

    def wrap(state, resp, *args):
        assert args == wrapper_args
        nonlocal result_wrapped
        result_wrapped = True

    aio_connection.add_cleanup_callback(cleanup)

    await aio_connection(
        request,
        stub,
        _apis.SchemeService.DescribePath,
        wrap_result=wrap,
        wrap_args=wrapper_args,
    )

    await aio_connection.close()

    assert result_wrapped and cleanup_callback_called
